from __future__ import annotations

import pytest

from pyaerocom.aeroval.experiment_output import ExperimentOutput
from pyaerocom.aeroval.experiment_processor import ExperimentProcessor
from pyaerocom.aeroval import EvalSetup


@pytest.mark.parametrize("cfg", ["cfgexp1"])
def test_ExperimentProcessor___init__(eval_config: dict):
    cfg = EvalSetup(**eval_config)
    proc = ExperimentProcessor(cfg)
    assert isinstance(proc.cfg, EvalSetup)
    assert isinstance(proc.exp_output, ExperimentOutput)


@pytest.fixture
def processor(eval_config: dict) -> ExperimentProcessor:
    """ExperimentProcessor instance without experiment data"""
    setup = EvalSetup(**eval_config)
    proc = ExperimentProcessor(setup)
    proc.exp_output.delete_experiment_data(also_coldata=True)
    return proc


@pytest.mark.parametrize("cfg", ["cfgexp1", "cfgexp2", "cfgexp3", "cfgexp4", "cfgexp5"])
def test_ExperimentProcessor_run(processor: ExperimentProcessor):
    processor.run()


# Temporary until ObsCollection implemented similarly then can run same test
@pytest.mark.parametrize(
    "cfg,kwargs,error",
    [
        (
            "cfgexp2",
            dict(model_name="BLA"),
            "'No matches could be found that match input BLA'",
        ),
        (
            "cfgexp2",
            dict(obs_name="BLUB"),
            "'No matches could be found that match input BLUB'",
        ),
    ],
)
def test_ExperimentProcessor_run_error_obs_name(
    processor: ExperimentProcessor, kwargs: dict, error: str
):
    with pytest.raises(KeyError) as e:
        processor.run(**kwargs)
    assert str(e.value) == error


@pytest.mark.parametrize(
    "cfg,kwargs,error",
    [
        ("cfgexp2", dict(var_list="BLUB"), "var_list %s and obs_vars %s mismatch."),
    ],
)
def test_ExperimentProcessor_catch_wrong_var_list(
    processor: ExperimentProcessor, kwargs: dict, error: str, caplog
):
    processor.run(**kwargs)
    assert any([error in str(record) for record in caplog.records])


@pytest.mark.parametrize("cfg", ["cfgexp2"])
def test_rerun_by_variable(eval_config: dict):
    # Add new obs network & var to the configuration for test
    eval_config["obs_cfg"]["AERONET-Inv"] = dict(
        obs_id="AeronetInvV3L2Subset.daily",
        obs_vars=("abs550aer",),
        obs_vert_type="Column",
    )
    eval_config["var_order_menu"] = ("od550aer", "abs550aer")
    setup = EvalSetup(**eval_config)
    processor = ExperimentProcessor(setup)
    processor.run()  # Initial full run
    assert (
        processor.exp_output.results_available
    ), "Experiment data should be available after initial run"
    assert "od550aer" in str(
        processor.exp_output
    ), "od550aer should be present in results after inital run"
    assert "abs550aer" in str(
        processor.exp_output
    ), "abs550aer should be present in results after initial run"
    processor.exp_output.delete_experiment_data(also_coldata=True)  # Clear previous data
    assert not processor.exp_output.results_available, "Experiment data should be cleared"
    # Rerun with the od550aer
    processor.run(var_list="od550aer")
    assert (
        processor.exp_output.results_available
    ), "Experiment data should be available after rerun of od550aer"
    assert "od550aer" in str(processor.exp_output)
    assert "abs550aer" not in str(
        processor.exp_output
    ), "abs550aer should not be present in results after rerun of just od550aer"
    # Rerun with the abs550aer
    processor.run(var_list="abs550aer")
    assert (
        processor.exp_output.results_available
    ), "Experiment data should be available after rerun of abs550aer"
    assert "od550aer" in str(
        processor.exp_output
    ), "od550aer should be present in results after rerun with abs550aer"
    assert "abs550aer" in str(
        processor.exp_output
    ), "abs550aer should be present in results after rerun abs550aer"


@pytest.mark.parametrize("cfg", ["cfgexp1"])
def test_rerun_by_model(eval_config: dict):
    setup = EvalSetup(**eval_config)
    processor = ExperimentProcessor(setup)
    processor.run()  # Initial full run
    assert (
        processor.exp_output.results_available
    ), "Experiment data should be available after initial run"
    assert "TM5-AP3-CTRL" in str(
        processor.exp_output
    ), "TM5-AP3-CTRL model should be present in results after initial run"
    processor.exp_output.delete_experiment_data(also_coldata=True)
    assert not processor.exp_output.results_available, "Experiment data should be cleared"
    assert "TM5-AP3-CTRL" not in str(
        processor.exp_output
    ), "TM5-AP3-CTRL model should not be present in results after clearing"
    # Rerun with the TM5-AP3-CTRL model
    processor.run(model_name="TM5-AP3-CTRL")
    assert (
        processor.exp_output.results_available
    ), "Experiment data should be available after rerun"
    assert "TM5-AP3-CTRL" in str(
        processor.exp_output
    ), "TM5-AP3-CTRL model should be present in results after rerun"
    # Rerun with a non-existing model
    with pytest.raises(KeyError) as e:
        processor.run(model_name="NonExistingModel")
    assert "No matches could be found that match input NonExistingModel" in str(
        e.value
    ), "Should raise KeyError for non-existing model"


@pytest.mark.parametrize("cfg", ["cfgexp4"])
def test_rerun_by_obs_network(eval_config: dict):
    # Ensure that the obs networks are set up correctly - want more than just superobs for this test
    eval_config["obs_cfg"]["AERONET-Sun"]["only_superobs"] = False  # Ensure superobs is not set
    eval_config["obs_cfg"]["AERONET-SDA"]["only_superobs"] = False
    setup = EvalSetup(**eval_config)
    processor = ExperimentProcessor(setup)
    processor.run()  # Initial full run
    assert (
        processor.exp_output.results_available
    ), "Experiment data should be available after initial run"
    assert "AERONET-Sun" in str(
        processor.exp_output
    ), "AERONET-Sun network should be present in results after initial run"
    assert "AERONET-SDA" in str(
        processor.exp_output
    ), "AERONET-SDA network should be present in results after initial run"
    assert "SDA-and-Sun" in str(
        processor.exp_output
    ), "SDA-and-Sun superobs network should be present in results after initial run"
    processor.exp_output.delete_experiment_data(also_coldata=True)  # Clear previous data
    assert not processor.exp_output.results_available, "Experiment data should be cleared"

    # Test rerun with single observation network
    processor.run(obs_name="AERONET-Sun")
    assert (
        processor.exp_output.results_available
    ), "Experiment data should be available after rerun of AERONET-Sun"
    assert "AERONET-Sun" in str(
        processor.exp_output
    ), "AERONET-Sun network should be present in results after rerun of AERONET-Sun"
    assert "AERONET-SDA" not in str(
        processor.exp_output
    ), "AERONET-SDA network should not be present in results after rerun of AERONET-Sun"
    assert "SDA-and-Sun" not in str(
        processor.exp_output
    ), "SDA-and-Sun superobs network should not be present in results after rerun of AERONET-Sun"
    processor.run(obs_name="AERONET-SDA")
    # Test that once a network has already been processed, another can be processed without clearing previous results
    assert (
        processor.exp_output.results_available
    ), "Experiment data should be available after rerun of AERONET-SDA"
    assert "AERONET-Sun" in str(
        processor.exp_output
    ), "AERONET-Sun network should be present in results after rerun of AERONET-SDA"
    assert "AERONET-SDA" in str(
        processor.exp_output
    ), "AERONET-SDA network should be present in results after rerun of AERONET-SDA"
    assert "SDA-and-Sun" not in str(
        processor.exp_output
    ), "SDA-and-Sun superobs network should not be present in results after rerun of AERONET-SDA"
    # Now test that the superobs can be re-added and that all previously processed networks are still present
    processor.run(obs_name="SDA-and-Sun")
    assert (
        processor.exp_output.results_available
    ), "Experiment data should be available after rerun of superobs SDA-and-Sun"
    assert "AERONET-Sun" in str(
        processor.exp_output
    ), "AERONET-Sun network should be present in results after rerun of superobs SDA-and-Sun"
    assert "AERONET-SDA" in str(
        processor.exp_output
    ), "AERONET-SDA network should be present in results after rerun of superobs SDA-and-Sun"
    assert (
        "SDA-and-Sun" in str(processor.exp_output)
    ), "SDA-and-Sun superobs observation network should be present in results after rerun of superobs SDA-and-Sun"



import iris
import numpy as np
import pandas as pd
import pytest
from cf_units import Unit

from pyaerocom import GriddedData, const, helpers, GriddedDataContainer
from pyaerocom.colocation.colocated_data import ColocatedData
from pyaerocom.colocation.colocation_utils import (
    _colocate_site_data_helper,
    _colocate_site_data_helper_timecol,
    _regrid_gridded,
    colocate_gridded_gridded,
    colocate_gridded_ungridded,
    _get_stat_data_vec,
)
from pyaerocom.config_reader import ALL_REGION_NAME
from pyaerocom.exceptions import UnresolvableTimeDefinitionError
from pyaerocom.io.mscw_ctm.reader import ReadMscwCtm
from tests.conftest import TEST_RTOL, need_iris_32
from tests.fixtures.stations import create_fake_station_data


def test__regrid_gridded(data_tm5):
    one_way = _regrid_gridded(data_tm5, "areaweighted", 5)
    another_way = _regrid_gridded(data_tm5, "areaweighted", dict(lon_res_deg=5, lat_res_deg=5))

    assert one_way.shape == another_way.shape


S1 = create_fake_station_data(
    "concpm10",
    {"concpm10": {"units": "ug m-3"}},
    10,
    "2010-01-01",
    "2010-12-31",
    "d",
    {"ts_type": "daily"},
)

S1["concpm10"][10:20] = np.nan

S2 = create_fake_station_data(
    "concpm10",
    {"concpm10": {"units": "ug m-3"}},
    10,
    "2010-01-01",
    "2010-12-31",
    "d",
    {"ts_type": "daily"},
)

S3 = create_fake_station_data(
    "concpm10",
    {"concpm10": {"units": "ug m-3"}},
    10,
    "2010-01-01",
    "2010-12-31",
    "13d",
    {"ts_type": "13daily"},
)
S3["concpm10"][1] = np.nan
S3["concpm10"][3] = np.nan

S4 = create_fake_station_data(
    "concpm10",
    {"concpm10": {"units": "ug m-3"}},
    10,
    "2010-01-03",
    "2011-12-31",
    "d",
    {"ts_type": "daily"},
)

S4["concpm10"][0:5] = range(5)


@pytest.mark.parametrize(
    "stat_data,stat_data_ref,var,var_ref,ts_type,resample_how,min_num_obs, use_climatology_ref,num_valid",
    [
        (
            S4,
            S3,
            "concpm10",
            "concpm10",
            "monthly",
            "mean",
            {"monthly": {"daily": 25}},
            False,
            10,
        ),
        (
            S3,
            S4,
            "concpm10",
            "concpm10",
            "monthly",
            "mean",
            {"monthly": {"daily": 25}},
            False,
            24,
        ),
        (S1, S2, "concpm10", "concpm10", "monthly", "mean", 25, {}, 12),
        (S2, S1, "concpm10", "concpm10", "monthly", "mean", 25, {}, 11),
    ],
)
def test__colocate_site_data_helper_timecol(
    stat_data,
    stat_data_ref,
    var,
    var_ref,
    ts_type,
    resample_how,
    min_num_obs,
    use_climatology_ref,
    num_valid,
):
    result = _colocate_site_data_helper_timecol(
        stat_data,
        stat_data_ref,
        var,
        var_ref,
        ts_type,
        resample_how,
        min_num_obs,
        use_climatology_ref,
    )

    assert isinstance(result, pd.DataFrame)
    assert result.data.isnull().sum() == result.ref.isnull().sum()
    assert len(result) - result.data.isnull().sum() == num_valid


def test__colocate_site_data_helper(aeronetsunv3lev2_subset):
    var = "od550aer"
    stat1 = aeronetsunv3lev2_subset.to_station_data(3, var)
    stat2 = aeronetsunv3lev2_subset.to_station_data(4, var)
    df = _colocate_site_data_helper(stat1, stat2, var, var, "daily", None, None, False)

    assert isinstance(df, pd.DataFrame)
    assert len(df) == 9483
    assert df["data"].mean() == pytest.approx(0.31171085422102346, rel=TEST_RTOL)
    assert df["ref"].mean() == pytest.approx(0.07752743643132792, rel=TEST_RTOL)


def test_colocate_gridded_ungridded_new_var(data_tm5, aeronetsunv3lev2_subset):
    data = data_tm5.copy()
    data.var_name = "od550bc"
    mdata = GriddedDataContainer("test_id")
    mdata.add_griddeddata(data)
    coldata = colocate_gridded_ungridded(mdata, aeronetsunv3lev2_subset, var_ref="od550aer")

    assert coldata.metadata["var_name"] == ["od550aer", "od550bc"]


@pytest.mark.parametrize(
    "addargs,ts_type,shape,obsmean,modmean",
    [
        (
            dict(
                filter_name=f"{ALL_REGION_NAME}-noMOUNTAINS",
                min_num_obs=const.OBS_MIN_NUM_RESAMPLE,
            ),
            "monthly",
            (2, 12, 8),
            0.315930,
            0.275671,
        ),
        (
            dict(filter_name=f"{ALL_REGION_NAME}-noMOUNTAINS"),
            "monthly",
            (2, 12, 8),
            0.316924,
            0.275671,
        ),
        (
            dict(
                filter_name=f"{ALL_REGION_NAME}-wMOUNTAINS",
                min_num_obs=const.OBS_MIN_NUM_RESAMPLE,
            ),
            "monthly",
            (2, 12, 11),
            0.269707,
            0.243861,
        ),
        pytest.param(
            dict(
                filter_name=f"{ALL_REGION_NAME}-noMOUNTAINS",
                regrid_res_deg=30,
                min_num_obs=const.OBS_MIN_NUM_RESAMPLE,
            ),
            "monthly",
            (2, 12, 8),
            0.31593,
            # 0.1797,
            0.169897,
            marks=[need_iris_32],
        ),
        (
            dict(filter_name=f"{ALL_REGION_NAME}-noMOUNTAINS", ts_type="yearly"),
            "yearly",
            (2, 1, 8),
            0.417676,
            0.275671,
        ),
    ],
)
def test_colocate_gridded_ungridded(
    data_tm5, aeronetsunv3lev2_subset, addargs, ts_type, shape, obsmean, modmean
):
    mdata = GriddedDataContainer("test_id")
    mdata.add_griddeddata(data_tm5)
    coldata = colocate_gridded_ungridded(mdata, aeronetsunv3lev2_subset, **addargs)

    assert isinstance(coldata, ColocatedData)
    assert coldata.ts_type == ts_type
    assert coldata.shape == shape

    assert np.nanmean(coldata.data.data[0]) == pytest.approx(obsmean, rel=TEST_RTOL)
    assert np.nanmean(coldata.data.data[1]) == pytest.approx(modmean, rel=TEST_RTOL)


def test_colocate_gridded_ungridded_split(cities_data, lcs_data):
    mdata_split = GriddedDataContainer("split")
    mdata_berlin = GriddedDataContainer("whole")

    for tile in cities_data["EMEP_split"]:
        mdata_split.add_griddeddata(tile)

    mdata_berlin.add_griddeddata(cities_data["EMEP"][1])

    coldata_split = colocate_gridded_ungridded(mdata_split, lcs_data)
    coldata_berlin = colocate_gridded_ungridded(mdata_berlin, lcs_data)

    assert len(coldata_split.data.station_name) == len(coldata_berlin.data.station_name)

    assert np.nanmean(coldata_split.data.data[0]) == np.nanmean(coldata_berlin.data.data[0])
    assert np.nanmean(coldata_split.data.data[1]) == np.nanmean(coldata_berlin.data.data[1])


def test__get_stat_data(cities_data, lcs_data):
    from pyaerocom.units.datetime import to_pandas_timestamp

    def proj(lat, lon):
        return (lon, lat)

    def coord_index(stationdata, points):
        lat = float(stationdata["latitude"])
        lon = float(stationdata["longitude"])

        dists2 = [(point[0] - lat) ** 2 + (point[1] - lon) ** 2 for point in points]
        return np.argmin(dists2)

    data_id = "test_id"
    var_ref = "concpm25"
    ts_type = "monthly"
    obs_start = to_pandas_timestamp("01-2019")
    obs_stop = to_pandas_timestamp("12-2019")
    mg = GriddedDataContainer(data_id)

    for gd in cities_data["EMEP"]:
        mg.add_griddeddata(gd)

    data_ref = lcs_data

    tiles, xranges, yranges = mg.get_tiles()

    data_ref = data_ref.filter_by_projection(proj, xranges, yranges)

    all_stats = data_ref.to_station_data_all(
        vars_to_convert=var_ref,
        start=obs_start,
        stop=obs_stop,
        by_station_name=True,
        ts_type_preferred=ts_type,
    )

    unsorted_obs_stat_data = all_stats["stats"]

    grid_stat_data, obs_stat_data = _get_stat_data_vec(
        obs_start,
        obs_stop,
        var_ref,
        tiles,
        xranges,
        yranges,
        proj,
        unsorted_obs_stat_data,
    )

    points_obs = [(stat["latitude"], stat["longitude"]) for stat in obs_stat_data]
    points_mod = [(stat["latitude"], stat["longitude"]) for stat in grid_stat_data]

    sorted_grid_stats = sorted(grid_stat_data, key=lambda x: coord_index(x, points_obs))
    sorted_obs_stats = sorted(unsorted_obs_stat_data, key=lambda x: coord_index(x, points_mod))

    assert sorted_grid_stats == grid_stat_data
    assert sorted_obs_stats == obs_stat_data
    assert sum(
        [i["latitude"] != j["latitude"] for i, j in zip(unsorted_obs_stat_data, obs_stat_data)]
    )
    assert sum(
        [i["longitude"] != j["longitude"] for i, j in zip(unsorted_obs_stat_data, obs_stat_data)]
    )


def test_colocate_gridded_ungridded_wstationtype(data_tm5, aeronetsunv3lev2_subset):
    fake_type = ["faketype"] * len(aeronetsunv3lev2_subset.station_name)
    mdata = GriddedDataContainer("test_id")
    mdata.add_griddeddata(data_tm5)
    for i, n in enumerate(fake_type):
        aeronetsunv3lev2_subset.metadata[i].update({"station_type": n})
    coldata = colocate_gridded_ungridded(
        mdata, aeronetsunv3lev2_subset, add_meta_keys=["station_type"]
    )
    assert all(typ == "faketype" for typ in coldata.coords["station_type"].values.tolist())


def test_colocate_gridded_ungridded_nonglobal(aeronetsunv3lev2_subset):
    times = [1, 2]
    time_unit = Unit("days since 2010-1-1 0:0:0")
    cubes = iris.cube.CubeList()

    for time in times:
        time_coord = iris.coords.DimCoord(time, units=time_unit, standard_name="time")
        cube = helpers.make_dummy_cube_latlon(
            lat_res_deg=1,
            lon_res_deg=1,
            lat_range=[30.05, 81.95],
            lon_range=[-29.5, 89.95],
        )
        cube.add_aux_coord(time_coord)
        cubes.append(cube)
    time_cube = cubes.merge_cube()
    gridded = GriddedData(time_cube)
    gridded.var_name = "od550aer"
    gridded.units = Unit("1")

    mdata = GriddedDataContainer("test_id")
    mdata.add_griddeddata(gridded)
    coldata = colocate_gridded_ungridded(mdata, aeronetsunv3lev2_subset, colocate_time=False)
    assert isinstance(coldata, ColocatedData)
    assert coldata.shape == (2, 2, 2)


def test_colocate_gridded_gridded_same_new_var(data_tm5):
    data = data_tm5.copy()

    data.var_name = "Blaaa"
    mdata = GriddedDataContainer("test", data)
    coldata = colocate_gridded_gridded(mdata, GriddedDataContainer("tast", data_tm5))

    assert coldata.metadata["var_name"] == ["od550aer", "Blaaa"]


def test_colocate_gridded_gridded_same(data_tm5):
    coldata = colocate_gridded_gridded(
        GriddedDataContainer("test1", data_tm5), GriddedDataContainer("test1", data_tm5)
    )

    assert isinstance(coldata, ColocatedData)
    stats = coldata.calc_statistics()
    # check mean value
    assert stats["data_mean"] == pytest.approx(0.09825691)
    # check that mean value is same as in input GriddedData object
    assert stats["data_mean"] == pytest.approx(data_tm5.mean(areaweighted=False))
    assert stats["refdata_mean"] == stats["data_mean"]
    assert stats["nmb"] == 0
    assert stats["mnmb"] == 0
    assert stats["R"] == pytest.approx(1, rel=1e-6)
    assert stats["R_spearman"] == pytest.approx(1, rel=1e-6)


@pytest.mark.xfail(raises=UnresolvableTimeDefinitionError)
def test_read_emep_colocate_emep_tm5(data_tm5, path_emep):
    reader = ReadMscwCtm(data_dir=path_emep["data_dir"])
    data_emep = reader.read_var("concpm10", ts_type="monthly")

    # Change units and year to match TM5 data
    data_emep.change_base_year(2010)
    data_emep.units = "1"

    col = colocate_gridded_gridded(data_emep, data_tm5)
    assert isinstance(col, ColocatedData)



import numpy as np
import logging
from copy import deepcopy
import iris

from .griddeddata import GriddedData
from pyaerocom.stationdata import StationData

from pyaerocom.io.gridded_reader import GriddedReader


logger = logging.getLogger(__name__)


class GriddedDataContainerException(Exception):
    pass


class GriddedDataContainer:
    def __init__(self, data_id: str, data: GriddedData | None = None):
        """
        Class for holding and working with multiple GriddedData objects.

        """
        self.data_id = data_id

        self.children: list[GriddedData] = []

        self.ndim = None

        self.from_files = []

        self.start = None
        self.stop = None

        self.proj_info = None

        self.units = None
        self.var_name = None
        self.ts_type = None

        self._lon_res = None
        self._lat_res = None

        self._lat_points = None
        self._lon_points = None

        if data is not None:
            self.add_griddeddata(data)

    def _initiate(self, data: GriddedData):
        """
        Starts the list of children, and defines all the important attributes for objects

        Parameters
        ----------

        data: GriddedData
        """
        self.proj_info = deepcopy(data.proj_info)
        self.start = data.start
        self.stop = data.stop

        self.var_name = data.var_name
        self.ts_type = data.ts_type
        self.units = data.units

        self.ndim = data.ndim

        try:
            self._lat_points = data.latitude.points
        except Exception:
            self._lat_points = None
            assert self.proj_info is not None

        try:
            self._lon_points = data.longitude.points
        except Exception:
            self._lon_points = None
            assert self.proj_info is not None

    def only_one_child(func):
        def check_child(self, *args, **kwargs):
            if len(self.children) > 1:
                raise NotImplementedError(
                    f"{func.__name__} is not implemented for cases with more than one child GriddedData"
                )
            else:
                return func(self, *args, **kwargs)

        return check_child

    def add_griddeddata(self, data: GriddedData):
        """
        Adds a new GriddedData to children after checking if this new GriddedData object is
        compatible with existing children

        Parameters
        ----------
        data: GriddedData

        Raises
        -------
        GriddedDataContainerException
            If proj_info, var_name, ts_type, units or ndims are different from the existing ones

        """
        if len(self.children) == 0:
            self._initiate(data)

        if type(data.proj_info) is not type(self.proj_info):
            raise GriddedDataContainerException(
                f"Proj info from added griddeddata and existing proj info do not have the same type: {type(data.proj_info), {type(self.proj_info)}}"
            )
        if data.proj_info is not None and self.proj_info is not None:
            if data.proj_info != self.proj_info:
                raise GriddedDataContainerException(
                    "Proj data of added griddeddata is different from the existing proj info"
                )

        if data.var_name != self.var_name:
            raise GriddedDataContainerException(
                f"var_name of added griddeddata {data.var_name} is different from the existing var name {self.var_name}"
            )

        if data.ts_type != self.ts_type:
            raise GriddedDataContainerException(
                f"ts_type of added griddeddata {data.ts_type} is different from the existing ts_type {self.ts_type}"
            )

        if data.units != self.units:
            raise GriddedDataContainerException(
                f"units of added griddeddata {data.units} is different from the existing units {self.units}"
            )

        if data.ndim != self.ndim:
            raise GriddedDataContainerException(
                f"ndim of added griddeddata {data.ndim} is different from the existing ndim {self.ndim}"
            )

        self.start = min(self.start, data.start)
        self.stop = max(self.stop, data.stop)

        self.from_files += data.from_files

        self.children.append(data)

    def read_data(
        self,
        readers: list[GriddedReader],
        var_name: str,
        start: str,
        stop: str,
        ts_type: str,
        vert_which: str,
        flex_ts_type: str,
        **kwargs,
    ) -> None:
        for reader in readers:
            data = reader.read_var(
                var_name,
                start=start,
                stop=stop,
                ts_type=ts_type,
                vert_which=vert_which,
                flex_ts_type=flex_ts_type,
                **kwargs,
            )

            self.add_griddeddata(data)

    def get_xyranges(self) -> tuple[list[tuple[float, float]]]:
        """
        Finds the max/min ranges for x and y of all children


        Returns
        -------
        tuple[list[tuple[float, float]]]
            Two list, one for xs and one for ys

        Raises
        --------
        GriddedDataContainerException
            If self has no proj_info
        ValueError
            If there is a child where x or y is not found

        """
        if self.proj_info is None:
            raise GriddedDataContainerException("X and Y cannot be found, since proj_info is None")

        xranges = []
        yranges = []
        for data in self.children:
            xrange, yrange = self._get_xyrange(data)
            if xrange is None or yrange is None:
                raise ValueError(
                    f"x/y axis not found in cube: {data.proj_info.x_axis}, {data.proj_info.y_axis}"
                )
            xranges.append(xrange)
            yranges.append(yrange)
        return xranges, yranges

    @staticmethod
    def _get_xyrange(child: GriddedData) -> tuple[tuple[float, float]] | tuple[None]:
        """
        Finds the range of the axis. Done by taking first and last point in each dimension.
        This might be the middle point of the bounding cells

        Parameters:
        child : GriddedData
            The GriddedData to find the ranges

        Returns:
        tuple[tuple[float, float]]
            Tuple of the ranges

        """
        xrange = None
        yrange = None
        for coord in child.cube.dim_coords:
            if coord.var_name == child.proj_info.x_axis:
                vals = coord.points
                xrange = (vals[0], vals[-1])

            if coord.var_name == child.proj_info.y_axis:
                vals = coord.points
                yrange = (vals[0], vals[-1])

        return xrange, yrange

    def get_tiles(
        self,
    ) -> tuple[list[GriddedData], list[tuple[float, float]], list[tuple[float, float]]]:
        """
        Makes a list of available tiles, and lists of x- and yranges for said tiles


        Returns:
        list[GriddedData]
            list of tiles

        tuple[tuple[float, float]]
            list of the xranges

        tuple[tuple[float, float]]
            list of the yranges

        """
        tiles = []
        xranges = []
        yranges = []
        for data in self.children:
            if self.proj_info is None:
                latitude = data.latitude.points
                longitude = data.longitude.points
                xrange = [np.min(longitude), np.max(longitude)]
                yrange = [np.min(latitude), np.max(latitude)]
            else:
                xrange, yrange = self._get_xyrange(data)

            if len(xrange) == 0 or len(yrange) == 0:
                logger.warning(f"Tile {data.from_files} with no x and y range found")
                continue

            tiles.append(data)
            xranges.append(xrange)
            yranges.append(yrange)

        if not (len(tiles) == len(xranges) and len(tiles) == len(yranges)):
            raise GriddedDataContainerException(
                "Something went wrong then getting tiles. Length of ranges were different"
            )
        return tiles, xranges, yranges

    @property
    @only_one_child
    def lat_res(self):
        return self.children[0].lat_res

    @property
    @only_one_child
    def lon_res(self):
        return self.children[0].lon_res

    @property
    @only_one_child
    def latitude_points(self):
        return self._lat_points

    @property
    @only_one_child
    def longitude_points(self):
        return self._lon_points

    @property
    def longitude_circular(self):
        return all([data.longitude.circular for data in self.children])

    @property
    def latitude_circular(self):
        return all([data.latitude.circular for data in self.children])

    @only_one_child
    def get_cube_data(self):
        return self.children[0].cube.data

    def get_cube_data_all(self) -> list:
        return [data.cube.data for data in self.children]

    # Methods which are simply applied to each of the children

    # TODO: Make sure that in the children loop, obj.children[i] is updated in the loops below (eventhough the inplace should be working)

    @property
    def time(self):
        if len(self.children) > 1:
            logger.warning("Be careful with using this function with more than one child")
            points = []
            metadata = self.children[0].time.metadata
            units = self.children[0].time.units
            for data in self.children:
                if metadata != data.time.metadata:
                    raise GriddedDataContainerException(
                        "Time from different children has different metadata"
                    )
                if units != data.time.units:
                    raise GriddedDataContainerException(
                        "Time from different children has different units"
                    )
                points += list(data.time.points)

            return iris.coords.DimCoord(np.array(points), var_name="time", **units, **metadata)

        return self.children[0].time

    @property
    def has_latlon_dims(self):
        return all([data.has_latlon_dims for data in self.children])

    @property
    @only_one_child
    def grid(self):
        return self.children[0].grid

    @property
    def data_revision(self):
        revision = ""
        for i, data in enumerate(self.children):
            revision += f"nr {i}: {data.data_revision}; "
        return revision

    def time_stamps(self):
        time_stamps = []
        for data in self.children:
            time_stamps += list(data.time_stamps())

        return np.sort(np.unique(np.array(time_stamps)))

    def _check_lonlat_bounds(self):
        for data in self.children:
            data._check_lonlat_bounds()

    def check_lon_circular(self):
        return all([data.check_lon_circular() for data in self.children])

    def extract_surface_level(self):
        """Extract surface level from 4D field"""
        return self.children[0].extract_surface_level()

    def to_time_series(
        self,
        sample_points=None,
        scheme="nearest",
        vert_scheme=None,
        add_meta=None,
        use_iris=False,
        **coords,
    ) -> list[StationData]:
        raise NotImplementedError(
            "to_time_series is not implemented for this container, due to problems with sorting returned stationdata (compared with the stations datas of the obs)"
        )

    def register_var_glob(self, delete_existing=True):  # pragma: no cover
        """
        Applies register_var_glob function to first child, and returns result.

        See GriddedData for more info on this function
        """
        return self.children[0].register_var_glob(delete_existing)

    def regrid(
        self, other=None, lat_res_deg=None, lon_res_deg=None, scheme="areaweighted", **kwargs
    ):
        """
        Applies regrid function to all children.

        See GriddedData for more info on this function
        """
        if self.proj_info is not None:
            raise GriddedDataContainerException(
                "Could not regrid, since the data is projected. Please use data with latlon instead"
            )
        for i, data in enumerate(self.children):
            self.children[i] = data.regrid(other, lat_res_deg, lon_res_deg, scheme, **kwargs)

        return self

    def filter_region(self, region_id, inplace=False, **kwargs):  # pragma: no cover
        """
        Applies filter_region function to all children.

        See GriddedData for more info on this function
        """
        obj = self if inplace else self.copy()
        for i, data in enumerate(obj.children):
            obj.children[i] = data.filter_region(region_id, inplace, **kwargs)

        return obj

    def check_dimcoords_tseries(self):  # pragma: no cover
        """
        Applies check_dimcoords_tseries function to all children.

        See GriddedData for more info on this function
        """
        for data in self.children:
            data.check_dimcoords_tseries()

    def check_unit(self):  # pragma: no cover
        """
        Applies check_unit function to all children.

        See GriddedData for more info on this function
        """

        return all([data.check_unit() for data in self.children])

    def convert_unit(self, new_unit: str, inplace: bool = True):  # pragma: no cover
        """
        Applies convert_unit function to all children.

        See GriddedData for more info on this function
        """

        obj = self if inplace else self.copy()
        for i, data in enumerate(obj.children):
            obj.children[i] = data.convert_unit(
                new_unit,
            )

        return obj

    def remove_outliers(self, low, high, inplace=True):  # pragma: no cover
        """
        Applies remove_outliers function to all children.

        See GriddedData for more info on this function
        """
        obj = self if inplace else self.copy()
        for i, data in enumerate(obj.children):
            obj.children[i] = data.remove_outliers(low, high, inplace=True)
        return obj

    def reorder_dimensions_tseries(self):  # pragma: no cover
        """
        Applies reorder_dimensions_tseries function to all children.

        See GriddedData for more info on this function
        """
        for data in self.children:
            data.reorder_dimensions_tseries()

    def crop(
        self, lon_range=None, lat_range=None, time_range=None, region=None
    ):  # pragma: no cover
        """
        Applies crop function to all children, and return self.

        See GriddedData for more info on this function
        """
        for i, data in enumerate(self.children):
            self.children[i] = data.crop(lon_range, lat_range, time_range, region)

        return self

    def resample_time(
        self, to_ts_type, how=None, min_num_obs=None, use_iris=False
    ):  # pragma: no cover
        """
        Applies resample_time function to all children, and return self.

        See GriddedData for more info on this function
        """
        for i, data in enumerate(self.children):
            self.children[i] = data.resample_time(to_ts_type, how, min_num_obs, use_iris)

        return self

    def filter_altitude(self, alt_range=None):  # pragma: no cover
        """
        Applies filter_altitude function to all children, and return self.

        See GriddedData for more info on this function
        """
        logger.warning(
            "Altitude filtering is not applied in GriddedDataContainer and will be skipped"
        )

        return self



"""
Helpers for conversion of ColocatedData to JSON files for web interface.
"""

import logging
from collections.abc import Callable
from copy import deepcopy
from dataclasses import dataclass
from datetime import datetime, timezone
from typing import Literal

import numpy as np
import pandas as pd
import xarray as xr

from pyaerocom import ColocatedData
from pyaerocom._warnings import ignore_warnings
from pyaerocom.aeroval.exceptions import ConfigError, TrendsError

# from pyaerocom.aeroval.experiment_output import ExperimentOutput
from pyaerocom.aeroval.fairmode_stats import fairmode_stats
from pyaerocom.aeroval.helpers import (
    _get_min_max_year_periods,
    _period_str_to_timeslice,
)
from pyaerocom.config_reader import ALL_REGION_NAME
from pyaerocom.exceptions import DataCoverageError, TemporalResolutionError
from pyaerocom.region import (
    Region,
    find_closest_region_coord,
    get_all_default_region_ids,
)
from pyaerocom.region_defs import HTAP_REGIONS_DEFAULT, OLD_AEROCOM_REGIONS, EU_CITIES_REGIONS
from pyaerocom.stats.stats import _init_stats_dummy, calculate_statistics
from pyaerocom.trends_engine import TrendsEngine
from pyaerocom.trends_helpers import (
    _get_season_from_months,
    _get_unique_seasons,
    _get_yearly,
    _init_trends_result_dict,
)
from pyaerocom.units.datetime import TsType

# from pyaerocom.aeroval.fairmode_statistics import FairmodeStatistics

logger = logging.getLogger(__name__)


def _prepare_regions_json_helper(region_ids):
    regborders, regs = {}, {}
    for regid in region_ids:
        reg = Region(regid)
        name = reg.name
        regs[name] = reg
        regborders[name] = rinfo = {}

        latr = reg.lat_range
        lonr = reg.lon_range
        if any(x is None for x in (latr, lonr)):
            raise ValueError(f"Lat / lon range missing for region {regid}")
        rinfo["minLat"] = latr[0]
        rinfo["maxLat"] = latr[1]
        rinfo["minLon"] = lonr[0]
        rinfo["maxLon"] = lonr[1]

    return (regborders, regs)


def _prepare_default_regions_json():
    return _prepare_regions_json_helper(get_all_default_region_ids())


def _prepare_aerocom_regions_json():
    return _prepare_regions_json_helper(OLD_AEROCOM_REGIONS)


def _prepare_htap_regions_json():
    return _prepare_regions_json_helper(HTAP_REGIONS_DEFAULT)


def _prepare_cities_regions_json():
    return _prepare_regions_json_helper(EU_CITIES_REGIONS)


def _prepare_country_regions(region_ids):
    regs = {}
    for regid in region_ids:
        reg = Region(regid)
        name = reg.name
        regs[name] = reg
    return regs


def init_regions_web(coldata, regions_how):
    regborders, regs = {}, {}
    regborders_default, regs_default = _prepare_default_regions_json()
    if regions_how == "default":
        regborders, regs = regborders_default, regs_default
    elif regions_how == "aerocom":
        regborders, regs = _prepare_aerocom_regions_json()
    elif regions_how == "htap":
        regborders[ALL_REGION_NAME] = regborders_default[ALL_REGION_NAME]
        regs[ALL_REGION_NAME] = regs_default[ALL_REGION_NAME]
        add_borders, add_regs = _prepare_htap_regions_json()
        regborders.update(add_borders)
        regs.update(add_regs)
    elif regions_how == "country":
        regborders[ALL_REGION_NAME] = regborders_default[ALL_REGION_NAME]
        regs[ALL_REGION_NAME] = regs_default[ALL_REGION_NAME]
        coldata.check_set_countries()
        regborders.update(coldata.get_country_codes())
        add_regs = _prepare_country_regions(coldata.get_country_codes().keys())
        regs.update(add_regs)
    elif regions_how == "cities":
        regborders[ALL_REGION_NAME] = regborders_default[ALL_REGION_NAME]
        regs[ALL_REGION_NAME] = regs_default[ALL_REGION_NAME]
        add_borders, add_regs = _prepare_cities_regions_json()
        regborders.update(add_borders)
        regs.update(add_regs)
    else:
        raise ValueError("Invalid input for regions_how", regions_how)

    regnames = {}
    for regname, reg in regs.items():
        regnames[reg.region_id] = regname
    return (regborders, regs, regnames)


def _init_meta_glob(coldata, **kwargs):
    meta = coldata.metadata

    # create metadata dictionary that is shared among all timeseries files
    meta_glob = {}
    NDSTR = "UNDEFINED"
    try:
        meta_glob["pyaerocom_version"] = meta["pyaerocom"]
    except KeyError:
        meta_glob["pyaerocom_version"] = NDSTR
    try:
        meta_glob["obs_var"] = meta["var_name"][0]
        meta_glob["mod_var"] = meta["var_name"][1]
    except KeyError:
        meta_glob["obs_var"] = NDSTR
        meta_glob["mod_var"] = NDSTR
    try:
        meta_glob["obs_unit"] = meta["var_units"][0]
        meta_glob["mod_unit"] = meta["var_units"][1]
    except KeyError:
        meta_glob["obs_unit"] = NDSTR
        meta_glob["mod_unit"] = NDSTR
    try:
        meta_glob["obs_freq_src"] = meta["ts_type_src"][0]
        meta_glob["mod_freq_src"] = meta["ts_type_src"][1]
    except KeyError:
        meta_glob["obs_freq_src"] = NDSTR
        meta_glob["mod_freq_src"] = NDSTR
    try:
        meta_glob["obs_revision"] = meta["revision_ref"]
    except KeyError:
        meta_glob["obs_revision"] = NDSTR
    meta_glob["processed_utc"] = datetime.now(timezone.utc).strftime("%Y-%m-%d %H:%M")
    meta_glob.update(kwargs)
    return meta_glob


def _init_ts_data(freqs):
    data = {}
    for freq in freqs:
        data[f"{freq}_date"] = []
        data[f"{freq}_obs"] = []
        data[f"{freq}_mod"] = []
    return data


def _create_diurnal_weekly_data_object(coldata, resolution):
    """
    Private helper functions that creates the data set containing all the
    weekly time series at the specified resolution. Called by
    _process_sites_weekly_ts. The returned xarray.Dataset contains a dummy time
    variable (currently not used) and the weekly time series as xarray.DataArray
    objects.

    Parameters
    ----------
    coldata : ColocatedData
        ColocatedData object colocated on hourly resolution.
    resolution : string
        String specifying the averaging window used to generate representative
        weekly time series with hourly resolution. Valid values are 'yearly'
        and  'seasonal'.

    Raises
    ------
    ValueError
        If an invalid resolution is given raise ValueError and print what
        was given.

    Returns
    -------
    rep_week_full_period : xarray.Dataset
        Contains the weekly time series as the variable 'rep_week'


    """
    import xarray as xr

    data_allyears = coldata.data

    yearkeys = list(data_allyears.groupby("time.year").groups)

    if resolution == "seasonal":
        seasons = ["DJF", "MAM", "JJA", "SON"]
    elif resolution == "yearly":
        seasons = ["year"]
    else:
        raise ValueError(f"Invalid resolution. Got {resolution}.")

    first = True
    for yr in yearkeys:
        data = data_allyears.where(data_allyears["time.year"] == yr, drop=True)
        for seas in seasons:
            rep_week_ds = xr.Dataset()
            if resolution == "seasonal":
                mon_slice = data.where(data["time.season"] == seas, drop=True)
            elif resolution == "yearly":
                mon_slice = data

            month_stamp = f"{seas}"

            for day in range(7):
                day_slice = mon_slice.where(mon_slice["time.dayofweek"] == day, drop=True)
                rep_day = day_slice.groupby("time.hour").mean(dim="time")
                rep_day["hour"] = rep_day.hour / 24 + day + 1
                if day == 0:
                    rep_week = rep_day
                else:
                    rep_week = xr.concat([rep_week, rep_day], dim="hour")

            rep_week = rep_week.rename({"hour": "dummy_time"})
            month_stamps = np.zeros(rep_week.dummy_time.shape, dtype="<U5")
            month_stamps[:] = month_stamp
            rep_week_ds["rep_week"] = rep_week
            rep_week_ds["month_stamp"] = (("dummy_time"), month_stamps)

            if seas in ["DJF", "year"]:
                rep_week_full_period = rep_week_ds
            else:
                rep_week_full_period = xr.concat([rep_week_full_period, rep_week_ds], dim="period")

        if first:
            output_array = rep_week_full_period
            # output_array = output_array.expand_dims({'year':[0]},0)
            first = False
        else:
            output_array = xr.concat([output_array, rep_week_full_period], dim="year")

    try:
        output_array.year
    except AttributeError:
        output_array = output_array.expand_dims({"year": [0]}, 0)

    output_array = output_array.assign(dict(year=np.array(yearkeys)))
    # output_array.year[:] = np.array(yearkeys)

    return output_array


def _get_period_keys(resolution: Literal["seasonal", "yearly"]):
    period_keys = dict(
        seasonal=["DJF", "MAM", "JJA", "SON"],
        yearly=["All"],
    )

    if resolution not in period_keys:
        raise ValueError(f"Unknown {resolution=}")

    return period_keys[resolution]


def _process_one_station_weekly(stat_name, i, repw_res, meta_glob, time):
    """
    Processes one station data set for all supported averaging windows into a
    dict of station time series data and metadata.

    Parameters
    ----------
    stat_name : string
        Name of station to process
    i : int
        Index of the station to process in the xarray.Datasets.
    repw_res : dict
        Dictionary of xarray.Datasets for each supported averaging window for
        the weekly time series
    meta_glob : TYPE
        Dictionary containing global metadata.
    time : list
        Time index.

    Returns
    -------
    ts_data : dict
        Dictionary of time series data and metadata for one station. Contains all
        resolutions/averaging windows.
    has_data : bool
        Set to false if all data is missing for a station.

    """
    has_data = False

    years = list(repw_res["seasonal"].year.values)
    yeardict = {}
    for year in years:
        yeardict[f"{year}"] = {}

    ts_data = {
        "time": time,
        "seasonal": {"obs": deepcopy(yeardict), "mod": deepcopy(yeardict)},
        "yearly": {"obs": deepcopy(yeardict), "mod": deepcopy(yeardict)},
    }
    ts_data["station_name"] = stat_name
    ts_data.update(meta_glob)

    for y, year in enumerate(years):
        for res, repw in repw_res.items():
            repw = repw.transpose("year", "period", "data_source", "dummy_time", "station_name")
            obs_vals = repw[y, :, 0, :, i]
            if (np.isnan(obs_vals)).all().values:
                continue
            has_data = True
            mod_vals = repw[y, :, 1, :, i]

            period_keys = _get_period_keys(res)
            for period_num, pk in enumerate(period_keys):
                ts_data[res]["obs"][f"{year}"][pk] = obs_vals.sel(
                    period=period_num
                ).values.tolist()
                ts_data[res]["mod"][f"{year}"][pk] = mod_vals.sel(
                    period=period_num
                ).values.tolist()
    return ts_data, has_data


def _process_weekly_object_to_station_time_series(repw_res, meta_glob):
    """
    Process the xarray.Datasets objects returned by _create_diurnal_weekly_data_object
    into a dictionary containing station time series data and metadata.

    Parameters
    ----------
    repw_res : dict
        Dictionary of xarray.Datasets for each supported averaging window for
        the weekly time series
    meta_glob : dict
        Dictionary containing global metadata.

    Returns
    -------
    ts_objs : list
        List of dicts containing station time series data and metadata.

    """
    ts_objs = []
    dc = 0
    time = (np.arange(168) / 24 + 1).round(4).tolist()
    for i, stat_name in enumerate(repw_res["seasonal"].station_name.values):
        ts_data, has_data = _process_one_station_weekly(stat_name, i, repw_res, meta_glob, time)

        if has_data:
            ts_objs.append(ts_data)
            dc += 1
    return ts_objs


def _process_weekly_object_to_country_time_series(repw_res, meta_glob, regions_how, region_ids):
    """
    Process the xarray.Dataset objects returned by _create_diurnal_weekly_data_object
    into a dictionary containing country average time series data and metadata.

    ToDo
    ----
    Implement regions_how for other values than country based

    Parameters
    ----------
    repw_res : dict
        Dictionary of xarray.Datasets for each supported averaging window for
        the weekly time series
    meta_glob : dict
        Dictionary containing global metadata.
    regions_how : string
        String describing how regions are to be processed. Regional time series
        are only calculated if regions_how = country.
    region_ids : dict
        Dict containing mapping of region IDs and names.

    Returns
    -------
    ts_objs_reg : list
        List of dicts containing station time series data and metadata.

    """
    ts_objs_reg = []
    time = (np.arange(168) / 24 + 1).round(4).tolist()

    years = list(repw_res["seasonal"].year.values)
    yeardict = {}
    for year in years:
        yeardict[f"{year}"] = {}

    if regions_how != "country":
        print("Regional diurnal cycles are only implemented for country regions, skipping...")
        ts_objs_reg = None
    else:
        for regid, regname in region_ids.items():
            ts_data = {
                "time": time,
                "seasonal": {"obs": deepcopy(yeardict), "mod": deepcopy(yeardict)},
                "yearly": {"obs": deepcopy(yeardict), "mod": deepcopy(yeardict)},
            }
            ts_data["station_name"] = regname
            ts_data.update(meta_glob)

            for y, year in enumerate(years):
                for res, repw in repw_res.items():
                    repw = repw.transpose(
                        "year", "period", "data_source", "dummy_time", "station_name"
                    )
                    if regid == ALL_REGION_NAME:
                        subset = repw
                    else:
                        subset = repw.where(repw.country == regid)

                    avg = subset.mean(dim="station_name")
                    obs_vals = avg[y, :, 0, :]
                    mod_vals = avg[y, :, 1, :]

                    period_keys = _get_period_keys(res)
                    for period_num, pk in enumerate(period_keys):
                        ts_data[res]["obs"][f"{year}"][pk] = obs_vals.sel(
                            period=period_num
                        ).values.tolist()
                        ts_data[res]["mod"][f"{year}"][pk] = mod_vals.sel(
                            period=period_num
                        ).values.tolist()

            ts_objs_reg.append(deepcopy(ts_data))
    return ts_objs_reg


def _process_sites_weekly_ts(coldata, regions_how, region_ids, meta_glob):
    """
    Private helper function to process ColocatedData objects into dictionaries
    containing representative weekly time series with hourly resolution.

    Processing the coloceted data object into a collection of representative
    weekly time series is done in the private function _create_diurnal_weekly_data_object.
    This object (an xarray.Dataset) is then further processed into two dictionaries
    containing station and regional time series respectively.

    Parameters
    ----------
    coldata : ColocatedData
        The colocated data to process.
    regions_how : string
        String describing how regions are to be processed. Regional time series
        are only calculated if regions_how = country.
    region_ids : dict
        Dict containing mapping of region IDs and names.
    meta_glob : dict
        Dictionary containing global metadata.

    Returns
    -------
    ts_objs : list
        List of dicts containing station time series data and metadata.
    ts_objs_reg : list
        List of dicts containing country time series data and metadata.

    """
    assert coldata.dims == ("data_source", "time", "station_name")

    repw_res = {
        "seasonal": _create_diurnal_weekly_data_object(coldata, "seasonal")["rep_week"],
        "yearly": _create_diurnal_weekly_data_object(coldata, "yearly")["rep_week"].expand_dims(
            "period", axis=0
        ),
    }

    ts_objs = _process_weekly_object_to_station_time_series(repw_res, meta_glob)
    ts_objs_reg = _process_weekly_object_to_country_time_series(
        repw_res, meta_glob, regions_how, region_ids
    )

    return (ts_objs, ts_objs_reg)


@dataclass
class CoordSiteArrays:
    station_names: list[str]
    station_types: list[str]
    lats: list[float]
    lons: list[float]
    alts: list[float]
    countries: list[str]
    jsdates: dict[str, list[str]]
    display_names: list[str | None] | None


def _init_site_coord_arrays(data) -> CoordSiteArrays:
    found = False
    jsdates = {}
    for freq, cd in data.items():
        if cd is None:
            continue
        elif not found:
            sites = cd.data.station_name.values
            if "station_type" in cd.data.coords:
                sites_types = cd.data.station_type.values
            else:
                sites_types = [""] * len(sites)
            lats = cd.data.latitude.values.astype(np.float64)
            lons = cd.data.longitude.values.astype(np.float64)
            if "altitude" in cd.data.coords:
                alts = cd.data.altitude.values.astype(np.float64)
            else:
                alts = [np.nan] * len(lats)
            if "country" in cd.data.coords:
                countries = cd.data.country.values
            else:
                countries = ["UNAVAIL"] * len(lats)
            if "station_display_name" in cd.data.coords:
                display_names = cd.data["station_display_name"].values
            else:
                display_names = None

            found = True
        else:
            assert all(cd.data.station_name.values == sites)
        jsdates[freq] = cd.data.jsdate.values.tolist()
    return CoordSiteArrays(
        station_names=sites,
        station_types=sites_types,
        lats=lats,
        lons=lons,
        alts=alts,
        countries=countries,
        jsdates=jsdates,
        display_names=display_names,
    )


def _get_stat_regions(lats, lons, regions, **kwargs):
    regions_how = kwargs.get("regions_how", None)
    regs = []
    for lat, lon in zip(lats, lons):
        reg = find_closest_region_coord(lat, lon, regions=regions, regions_how=regions_how)
        regs.append(reg)
    return regs


def _process_sites(data, regions, regions_how, meta_glob):
    freqs = list(data)
    # (sites, site_types, lats, lons, alts, countries, jsdates)
    coord_arrays = _init_site_coord_arrays(data)
    if regions_how == "country":
        regs = coord_arrays.countries
    elif regions_how == "htap":
        regs = _get_stat_regions(
            coord_arrays.lats, coord_arrays.lons, regions, regions_how=regions_how
        )
    else:
        regs = _get_stat_regions(coord_arrays.lats, coord_arrays.lons, regions)

    ts_objs = []
    site_indices = []
    map_meta = []

    for i, site in enumerate(coord_arrays.station_names):
        # init empty timeseries data object
        site_meta = {
            "station_name": str(site),
            "station_type": coord_arrays.station_types[i],
            "latitude": coord_arrays.lats[i],
            "longitude": coord_arrays.lons[i],
            "altitude": coord_arrays.alts[i],
        }
        if regions_how == "country":
            site_meta["region"] = [regs[i]]
        else:
            site_meta["region"] = regs[i]
        if coord_arrays.display_names is not None:
            display_name = coord_arrays.display_names[i]
            if display_name is not None:
                site_meta["station_display_name"] = display_name
        ts_data = _init_ts_data(freqs)
        ts_data.update(meta_glob)
        ts_data.update(site_meta)
        has_data = False
        for freq, cd in data.items():
            if cd is not None:
                assert cd.dims == ("data_source", "time", "station_name")
                sitedata = cd.data.data[:, :, i]
                if np.all(np.isnan(sitedata)):
                    # skip this site, all is NaN
                    continue
                ts_data[f"{freq}_date"] = coord_arrays.jsdates[freq]
                ts_data[f"{freq}_obs"] = sitedata[0].tolist()
                ts_data[f"{freq}_mod"] = sitedata[1].tolist()
                has_data = True
        if has_data:  # site is valid
            # register ts_data
            ts_objs.append(ts_data)
            # remember site indices in data for faster processing of statistics
            # below.
            site_indices.append(i)
            # init map data for each valid site
            map_meta.append({**site_meta})

    return (ts_objs, map_meta, site_indices)


def _get_statistics(obs_vals, mod_vals, min_num, drop_stats):
    stats = calculate_statistics(mod_vals, obs_vals, min_num_valid=min_num, drop_stats=drop_stats)
    return _prep_stats_json(stats)


"""
TODO: Make the regional trends by finding all the stations with the correct
amount of data for a certain period and season -> Calculate trends for the average
of those stations

Almost the same as the _remove_less_covered function, but check for each season and period
"""


def _make_regional_trends(
    data: ColocatedData,
    trends_min_yrs: int,
    per: str,
    freq: str,
    season: str,
    regid: str,
    use_country: bool = True,
    median: bool = False,
) -> tuple[dict, dict, bool]:
    (start, stop) = _get_min_max_year_periods([per])
    (start, stop) = (start.year, stop.year)
    data = data.copy()

    station_mod_trends = []
    station_obs_trends = []

    trends_successful = True

    subset_time_series = data.get_regional_timeseries(
        regid,
        check_country_meta=use_country,
    )

    data = data.filter_region(
        regid,
        inplace=False,
        check_country_meta=use_country,
    )
    nb_stations = len(data.data.data[0, 0, :])

    for i in range(nb_stations):
        if stop - start >= trends_min_yrs:
            try:
                obs_vals = data.data.data[0, :, i]
                mod_vals = data.data.data[1, :, i]

                time = data.data.time.values
                (obs_trend, mod_trend) = _make_trends(
                    obs_vals,
                    mod_vals,
                    time,
                    freq,
                    season,
                    start,
                    stop,
                    trends_min_yrs,
                )
                if None not in obs_trend.values() and None not in mod_trend.values():
                    station_mod_trends.append(mod_trend)
                    station_obs_trends.append(obs_trend)
            except TrendsError as e:
                msg = f"Failed to calculate trends, and will skip. This was due to {e}"
                logger.info(msg)

    if len(station_obs_trends) == 0 or len(station_mod_trends) == 0:
        trends_successful = False
        obs_trend = {}
        mod_trend = {}
    else:
        obs_trend = _combine_regional_trends(
            station_obs_trends,
            start,
            stop,
            subset_time_series["obs"],
            median=median,
        )
        mod_trend = _combine_regional_trends(
            station_mod_trends,
            start,
            stop,
            subset_time_series["mod"],
            median=median,
        )

        obs_trend["data"] = obs_trend["data"].to_json()
        mod_trend["data"] = mod_trend["data"].to_json()

        obs_trend["map_var"] = f"slp_{start}"
        mod_trend["map_var"] = f"slp_{start}"

    return obs_trend, mod_trend, trends_successful


def _combine_regional_trends(
    trends: list[dict],
    start_yr: int | float,
    stop_yr: int | float,
    time_series: pd.Series,
    median: bool = False,
) -> dict:
    result = _init_trends_result_dict(start_yr)

    if median:
        avg_func = np.nanmedian
    else:
        avg_func = np.nanmean

    default_trend = trends[0]

    result["period"] = default_trend["period"]
    result["season"] = default_trend["season"]

    result["data"] = _get_yearly(time_series, default_trend["season"], start_yr).loc[
        str(start_yr) : str(stop_yr)
    ]

    result["n"] = _combine_statistic(trends, "n", np.nanmax)
    result["y_mean"] = _combine_statistic(trends, "y_mean", avg_func)
    result["y_min"] = _combine_statistic(trends, "y_min", np.nanmin)
    result["y_max"] = _combine_statistic(trends, "y_max", np.nanmax)

    # result["pval"] = _fishers_pval(trends)
    result["pval"] = _harmonic_mean_pval(trends)

    result["m"] = _combine_statistic(trends, "m", avg_func)
    result["m_err"] = _get_trend_error(trends, "m")
    result["yoffs"] = _combine_statistic(trends, "yoffs", avg_func)

    result["slp"] = _combine_statistic(trends, "slp", avg_func)
    result["slp_err"] = _get_trend_error(trends, "slp")
    result["reg0"] = _combine_statistic(trends, "reg0", avg_func)

    result[f"slp_{start_yr}"] = _combine_statistic(trends, f"slp_{start_yr}", avg_func)
    result[f"slp_{start_yr}_err"] = _get_trend_error(trends, f"slp_{start_yr}")
    result[f"reg0_{start_yr}"] = _combine_statistic(trends, f"reg0_{start_yr}", avg_func)

    return result


def _combine_statistic(data: list, key: str, func: Callable) -> float:
    cleaned = np.array([i[key] for i in data])
    return float(func(cleaned))


def _get_trend_error(trends: list, key: str) -> float:
    cleaned = np.array([i[key] for i in trends])
    return float(np.nanstd(cleaned))


def _fishers_pval(trends: list) -> float:
    """
    Quick implementation of Fisher's method for combining p values
    Is most probably wrong
    """
    pvals = np.array([i["pval"] for i in trends])
    x = -2.0 * np.nansum(np.log(pvals))
    return float(1 - x)


def _harmonic_mean_pval(trends: list, weights: np.ndarray | None = None) -> float:
    """
    Alternative way of combining pvals
    """
    pvals = np.array([i["pval"] for i in trends])
    L = len(pvals)
    if weights is not None and len(weights) != L:
        raise ValueError(
            f"Weights must be the same length as the number of pvals: {len(weights)} != {L}"
        )

    if weights is None:
        weights = np.array([1.0 / L for i in range(L)])

    har_mean = np.sum(weights) / np.sum(weights / pvals)

    return har_mean


def process_trends(
    subset: ColocatedData,
    trends_min_yrs: int,
    season: str,
    per: str,
    avg_over_trends: bool,
    regid: str,
    use_country: bool,
    freq: str,
    use_weights: bool,
) -> dict:
    # subset = _select_period_season_coldata(coldata, per, season, use_meteorological_seasons)
    stats = {}
    trends_successful = False
    mean_trends_successful = False
    median_trends_successful = False

    if avg_over_trends:
        (
            mean_obs_trend,
            mean_mod_trend,
            mean_trends_successful,
        ) = _make_regional_trends(
            subset,
            trends_min_yrs,
            per,
            freq,
            season,
            regid,
            use_country,
        )

        (
            median_obs_trend,
            median_mod_trend,
            median_trends_successful,
        ) = _make_regional_trends(
            subset,
            trends_min_yrs,
            per,
            freq,
            season,
            regid,
            use_country,
            median=True,
        )

    # Calculates the start and stop years. min_yrs have a test value of 7 years. Should be set in cfg
    (start, stop) = _get_min_max_year_periods([per])
    (start, stop) = (start.year, stop.year)

    if stop - start >= trends_min_yrs:
        try:
            subset_time_series = subset.get_regional_timeseries(
                regid, check_country_meta=use_country
            )

            (obs_trend, mod_trend) = _make_trends_from_timeseries(
                subset_time_series["obs"],
                subset_time_series["mod"],
                freq,
                season,
                start,
                stop,
                trends_min_yrs,
            )

            trends_successful = True
        except TrendsError as e:
            msg = f"Failed to calculate trends, and will skip. This was due to {e}"
            logger.info(msg)

    if trends_successful:
        # The whole trends dicts are placed in the stats dict
        stats["obs_trend"] = obs_trend
        stats["mod_trend"] = mod_trend

    if mean_trends_successful and avg_over_trends:
        # The whole trends dicts are placed in the stats dict
        stats["obs_mean_trend"] = mean_obs_trend
        stats["mod_mean_trend"] = mean_mod_trend

    if median_trends_successful and avg_over_trends:
        # The whole trends dicts are placed in the stats dict
        stats["obs_median_trend"] = median_obs_trend
        stats["mod_median_trend"] = median_mod_trend

    return stats


def _make_trends_from_timeseries(obs, mod, freq, season, start, stop, min_yrs):
    """
    Function for generating trends from timeseries

    Includes formatting in a way
    that can be serialized to json. A key, map_var, is added
    for use in the web interface.

    Parameters
    ----------
    obs     : pd.Series
        Time series of the obs
    mod     : pd.Series
        Time series of the mod
    freq    : str
        Frequency for the trends, either monthly or yearly
    season  : str
        Seasons used for the trends
    start   : int
        Start year
    stop    : int
        Stop year
    min_yrs : int
        Minimal number of years for the calculation of the trends

    Raises
    ------
    TrendsError
        If stop - start is smaller than min_yrs

    AeroValError
        If the trend engine returns None

    Returns
    ------
    (dict, dict)
        Dicts consisting of the trends data for the obs and mod
    """

    if stop - start < min_yrs:
        raise TrendsError(f"min_yrs ({min_yrs}) larger than time between start and stop")

    te = TrendsEngine

    # The model and observation data are made to pandas times series
    obs_trend_series = obs
    mod_trend_series = mod

    # Translate season to names used in trends_helpers.py. Should be handled there instead!
    season = _get_season_from_months(season)

    # Trends are calculated
    obs_trend = te.compute_trend(obs_trend_series, freq, start, stop, min_yrs, season)
    mod_trend = te.compute_trend(mod_trend_series, freq, start, stop, min_yrs, season)

    # Makes pd.Series serializable
    if obs_trend["data"] is None or mod_trend["data"] is None:
        raise TrendsError("Trends came back as None", obs_trend["data"], mod_trend["data"])

    obs_trend["data"] = obs_trend["data"].to_json()
    mod_trend["data"] = mod_trend["data"].to_json()

    obs_trend["map_var"] = f"slp_{start}"
    mod_trend["map_var"] = f"slp_{start}"

    return obs_trend, mod_trend


def _make_trends(obs_vals, mod_vals, time, freq, season, start, stop, min_yrs):
    """
    Function for generating trends from lists of observations

    This will calculate pandas time series
    from the lists and use that to calculate trends

    Parameters
    ----------
    obs     : list
        Time series of the obs
    mod     : list
        Time series of the mod
    freq    : str
        Frequency for the trends, either monthly or yearly
    season  : str
        Seasons used for the trends
    start   : int
        Start year
    stop    : int
        Stop year
    min_yrs : int
        Minimal number of years for the calculation of the trends

    Raises
    ------
    TrendsError
        If stop - start is smaller than min_yrs

    AeroValError
        If the trend engine returns None

    Returns
    ------
    (dict, dict)
        Dicts consisting of the trends data for the obs and mod
    """

    # The model and observation data are made to pandas times series
    obs_trend_series = pd.Series(obs_vals, time)
    mod_trend_series = pd.Series(mod_vals, time)

    (obs_trend, mod_trend) = _make_trends_from_timeseries(
        obs_trend_series, mod_trend_series, freq, season, start, stop, min_yrs
    )

    return obs_trend, mod_trend


def _process_map_and_scat(
    data: dict[str, ColocatedData | None],
    map_data: list[dict],
    site_indices: list[int],
    periods: list[str],
    scatter_freq: str,
    min_num: int,
    seasons: list[str],
    add_trends: bool,
    trends_min_yrs: int,
    avg_over_trends: bool,
    use_fairmode: bool,
    obs_var: str,
    drop_stats: tuple,
    use_meteorological_seasons: bool,
):
    stats_dummy = _init_stats_dummy(drop_stats=drop_stats)
    scat_data = {}
    scat_dummy = [np.nan]
    new_map_data: list[dict] = []
    for freq, cd in data.items():
        for per in periods:
            for season in seasons:
                use_dummy = True
                if cd is not None:
                    try:
                        subset = _select_period_season_coldata(
                            cd, per, season, use_meteorological_seasons
                        )
                        jsdate = subset.data.jsdate.values.tolist()
                        use_dummy = False
                    except (DataCoverageError, TemporalResolutionError):
                        pass

                for i, map_stat in zip(site_indices, map_data):
                    if freq not in map_stat:
                        map_stat[freq] = {}

                    if use_dummy:
                        stats = stats_dummy
                    else:
                        obs_vals = subset.data.data[0, :, i]
                        mod_vals = subset.data.data[1, :, i]
                        stats = _get_statistics(
                            obs_vals, mod_vals, min_num=min_num, drop_stats=drop_stats
                        )

                        if use_fairmode and freq != "yearly" and not np.isnan(obs_vals).all():
                            stats["mb"] = np.nanmean(mod_vals - obs_vals)

                            stats["fairmode"] = fairmode_stats(obs_var, stats, freq)

                        #  Code for the calculation of trends
                        if add_trends and freq != "daily":
                            (start, stop) = _get_min_max_year_periods([per])
                            (start, stop) = (start.year, stop.year)

                            if stop - start >= trends_min_yrs:
                                try:
                                    time = subset.data.time.values
                                    (obs_trend, mod_trend) = _make_trends(
                                        obs_vals,
                                        mod_vals,
                                        time,
                                        freq,
                                        season,
                                        start,
                                        stop,
                                        trends_min_yrs,
                                    )

                                    # The whole trends dicts are placed in the stats dict
                                    stats["obs_trend"] = obs_trend
                                    stats["mod_trend"] = mod_trend
                                    stats["units"] = subset.units[0]

                                    if avg_over_trends:
                                        stats["obs_mean_trend"] = obs_trend
                                        stats["mod_mean_trend"] = mod_trend

                                        stats["obs_median_trend"] = obs_trend
                                        stats["mod_median_trend"] = mod_trend

                                except TrendsError as e:
                                    msg = f"Failed to calculate trends, and will skip. This was due to {e}"
                                    logger.info(msg)

                    perstr = f"{per}-{season}"
                    map_stat[freq][perstr] = stats
                    if freq == scatter_freq:
                        # add only sites to scatter data that have data available
                        # in the lowest of the input resolutions (e.g. yearly)
                        site = map_stat["station_name"]
                        if site not in scat_data:
                            scat_data[site] = {}
                            scat_data[site]["latitude"] = map_stat["latitude"]
                            scat_data[site]["longitude"] = map_stat["longitude"]
                            scat_data[site]["altitude"] = map_stat["altitude"]
                            scat_data[site]["region"] = map_stat["region"]
                        if use_dummy:
                            obs = mod = jsdate = scat_dummy
                            units = None
                        else:
                            obs, mod = obs_vals.tolist(), mod_vals.tolist()
                            units = subset.units[0]
                        scat_data[site][perstr] = {
                            "obs": obs,
                            "mod": mod,
                            "date": jsdate,
                            "units": units,
                        }

                    new_map_data.append(map_stat)

    return (new_map_data, scat_data)


def _process_regional_timeseries(data, region_ids, regions_how, meta_glob):
    ts_objs = []
    freqs = list(data)
    check_countries = True if regions_how == "country" else False
    for regid, regname in region_ids.items():
        ts_data = _init_ts_data(freqs)
        ts_data["station_name"] = regname
        ts_data.update(meta_glob)

        for freq, cd in data.items():
            if cd is None:
                continue
            jsfreq = cd.data.jsdate.values.tolist()
            try:
                subset = cd.filter_region(regid, inplace=False, check_country_meta=check_countries)
            except DataCoverageError:
                logger.info(f"no data in {regid} ({freq}) to compute regional timeseries")
                ts_data[f"{freq}_date"] = jsfreq
                ts_data[f"{freq}_obs"] = [np.nan] * len(jsfreq)
                ts_data[f"{freq}_mod"] = [np.nan] * len(jsfreq)
                continue

            if subset.has_latlon_dims:
                avg = subset.data.mean(dim=("latitude", "longitude"))
            else:
                avg = subset.data.mean(dim="station_name")
            obs_vals = avg[0].data.tolist()
            mod_vals = avg[1].data.tolist()
            ts_data[f"{freq}_date"] = jsfreq
            ts_data[f"{freq}_obs"] = obs_vals
            ts_data[f"{freq}_mod"] = mod_vals

        ts_objs.append(ts_data)
    return ts_objs


def _apply_annual_constraint_helper(coldata, yearly):
    arr_yr = yearly.data
    yrs_cd = coldata.data.time.dt.year
    yrs_avail = arr_yr.time.dt.year
    obs_allyrs = arr_yr[0]
    for i, yr in enumerate(yrs_avail):
        obs_yr = obs_allyrs[i]
        nan_sites_yr = obs_yr.isnull()
        if not nan_sites_yr.any():
            continue
        scond = nan_sites_yr.data
        tcond = (yrs_cd == yr).data
        # workaround since numpy sometimes throws IndexError if tcond and
        # scond are attempted to be applied directly via
        # coldata.data.data[:,tcond, scond] = np.nan
        tsel = coldata.data.data[:, tcond]
        tsel[:, :, scond] = np.nan
        coldata.data.data[:, tcond] = tsel

    return coldata


def _apply_annual_constraint(data):
    """
    Apply annual filter to data

    Parameters
    ----------
    data : dict
        keys are frequencies, values are corresponding
        instances of `ColocatedData` in that resolution, or None, if colocated
        data is not available in that resolution (see also
        :func:`_init_data_default_frequencies`).

    Raises
    ------
    ConfigError
        If colocated data in yearly resolution is not available in input data

    Returns
    -------
    output : dict
        like input `data` but with annual constrained applied.

    """
    output = {}
    if "yearly" not in data or data["yearly"] is None:
        raise ConfigError(
            "Cannot apply annual_stats_constrained option. "
            'Please add "yearly" in your setup (see attribute '
            '"statistics_json" in AerocomEvaluation class)'
        )
    yearly = data["yearly"]
    for tst, cd in data.items():
        if cd is None:
            output[tst] = None
        else:
            output[tst] = _apply_annual_constraint_helper(cd, yearly)
    return output


def _prep_stats_json(stats):
    for k, v in stats.items():
        try:
            stats[k] = np.float64(v)  # for json encoder...
        except Exception:
            # value is str (e.g. for weighted stats)
            # 'NOTE': 'Weights were not applied to FGE and kendall and spearman corr (not implemented)'
            stats[k] = v
    return stats


def _get_extended_stats(coldata, use_weights, drop_stats):
    stats = coldata.calc_statistics(use_area_weights=use_weights, drop_stats=drop_stats)

    # Removes the spatial median and temporal mean (see mails between Hilde, Jonas, Augustin and Daniel from 27.09.21)
    # (stats['R_spatial_mean'],
    #  stats['R_spatial_median']) = _calc_spatial_corr(coldata, use_weights)

    # (stats['R_temporal_mean'],
    #  stats['R_temporal_median']) = _calc_temporal_corr(coldata)

    (stats["R_spatial_mean"], _) = _calc_spatial_corr(coldata, use_weights)

    (_, stats["R_temporal_median"]) = _calc_temporal_corr(coldata)

    return _prep_stats_json(stats)


def _calc_spatial_corr(coldata, use_weights):
    """
    Compute spatial correlation both for median and mean aggregation

    Parameters
    ----------
    coldata : ColocatedData
        Input data.
    use_weights : bool
        Apply area weights or not.

    Returns
    -------
    float
        mean spatial correlation
    float
        median spatial correlation

    """
    return (
        coldata.calc_spatial_statistics(aggr="mean", use_area_weights=use_weights)["R"],
        coldata.calc_spatial_statistics(aggr="median", use_area_weights=use_weights)["R"],
    )


def _calc_temporal_corr(coldata):
    """
    Compute temporal correlation both for median and mean aggregation

    Parameters
    ----------
    coldata : ColocatedData
        Input data.

    Returns
    -------
    float
        mean temporal correlation
    float
        median temporal correlation

    """
    if len(coldata.time) < 3:
        return np.nan, np.nan
    elif coldata.has_latlon_dims:
        coldata = coldata.flatten_latlondim_station_name()

    # Use only sites that contain at least 3 valid data points (otherwise
    # correlation will be 1).
    obs_ok = coldata.data[0].count(dim="time") > 2
    arr = []
    arr.append(coldata.data[0].where(obs_ok, drop=True))
    arr.append(coldata.data[1].where(obs_ok, drop=True))

    if np.prod(arr[0].shape) == 0 or np.prod(arr[1].shape) == 0:
        return np.nan, np.nan
    corr_time = xr.corr(arr[1], arr[0], dim="time")
    with ignore_warnings(RuntimeWarning, "Mean of empty slice", "All-NaN slice encountered"):
        return (np.nanmean(corr_time.data), np.nanmedian(corr_time.data))


def _select_period_season_coldata(
    coldata: ColocatedData, period: str, season: str, use_meteorological_seasons: bool
):
    tslice = _period_str_to_timeslice(period)
    if use_meteorological_seasons and len(period) == 4 and season == "DJF":
        # relevant only for single years and DJF
        # for period = '2022' tslice needs to be slice('2021-12','2022-11')
        yeardt = datetime.strptime(period, "%Y")
        tslice = slice(f"{yeardt.year - 1}-12", f"{yeardt.year}-11")
        logger.debug(f"Using meteorological year slicing for {period}: {tslice}")
    # expensive, try use solution with numpy indexing directly...
    # also, keep an eye on: https://github.com/pydata/xarray/issues/2799
    arr = coldata.data.sel(time=tslice)
    if len(arr.time) == 0:
        raise DataCoverageError(f"No data available in period {period}")
    if season != "all":
        if season not in arr.season:
            raise DataCoverageError(f"No data available in {season} in period {period}")
        elif TsType(coldata.ts_type) < "monthly":
            raise TemporalResolutionError(
                "Season selection is only available for monthly or higher resolution data"
            )
        mask = arr["season"] == season
        arr = arr.sel(time=arr["time"][mask])

    return ColocatedData(data=arr)


def _process_heatmap_data(
    data,
    region_ids,
    use_weights,
    drop_stats,
    use_country,
    meta_glob,
    periods,
    seasons,
    add_trends,
    trends_min_yrs,
    avg_over_trends,
    use_meteorological_seasons,
):
    output = {}
    stats_dummy = _init_stats_dummy(drop_stats=drop_stats)
    for freq, coldata in data.items():
        output[freq] = hm_freq = {}
        for regid, regname in region_ids.items():
            hm_freq[regname] = {}
            for per in periods:
                for season in seasons:
                    use_dummy = coldata is None
                    perstr = f"{per}-{season}"
                    if use_dummy:
                        stats = stats_dummy
                    else:
                        try:
                            subset = _select_period_season_coldata(
                                coldata, per, season, use_meteorological_seasons
                            )

                            if add_trends and freq != "daily":
                                trend_stats = process_trends(
                                    subset,
                                    trends_min_yrs,
                                    season,
                                    per,
                                    avg_over_trends,
                                    regid,
                                    use_country,
                                    freq,
                                    use_weights,
                                )

                            subset = subset.filter_region(
                                region_id=regid, check_country_meta=use_country
                            )

                            stats = _get_extended_stats(subset, use_weights, drop_stats)

                            if add_trends and freq != "daily":
                                stats.update(**trend_stats)

                        except (DataCoverageError, TemporalResolutionError):
                            stats = stats_dummy

                    hm_freq[regname][perstr] = stats

    return output


def _map_indices(outer_idx, inner_idx):
    """
    Find index positions of inner array contained in outer array

    Note
    ----
    Both outer and inner index arrays must be strictly monotonous

    Parameters
    ----------
    outer_idx : np.ndarray
        numerical value array (e.g. numerical timestamps in yearly resolution
        from 2010-2020). Must be strictly monotonous.
    inner_idx : np.ndarray
        inner value array, must be contained in outer array, e.g.
        numerical timestamps in yearly resolution from 2012-2014, all values
        in this array must be contained

    Returns
    -------
    mappingg : np.ndarray
        same shape as `outer_idx` with values -1 where no matches are found
        and else, corresponding indices of `inner_idx`

    Example
    -------
    `outer_idx`: [2010, 2011, 2012, 2013, 2014, 2015]
    `inner_idx`: [2011, 2012]
    `mapping`  : [-1, 0, 1, -1, -1, -1]

    """
    inner_start, inner_stop = inner_idx[0], inner_idx[-1]
    mapping = np.ones_like(outer_idx) * -1
    count = 0
    indata = False
    for i, idx in enumerate(outer_idx):
        if inner_start == idx:
            indata = True
        elif inner_stop < idx:
            break
        if indata:
            mapping[i] = count
            # sanity checking, will fail, e.g. if input data is not monotonuos
            assert inner_idx[count] == idx
            count += 1
    return mapping.astype(int)


def _process_statistics_timeseries(
    data, freq, region_ids, use_weights, drop_stats, use_country, data_freq
):
    """
    Compute statistics timeseries for input data

    Parameters
    ----------
    data : dict
        dictionary containing colocated data object (values) in different
        temporal resolutions (keys).
    freq : str
        Output frequency (temporal resolution in which statistics timeseries
        if computed, AeroVal default is monthly)
    region_ids : dict
        Region IDs (keys) and corresponding names (values)
    use_weights : bool
        calculate statistics using area weights or not (only relevant for 4D
        colocated data with lat and lon dimension, e.g. from gridded / gridded
        co-location)
    use_country : bool
        Use countries for regional filtering.
    data_freq : str, optional
        Base frequency for computation of statistics (if None, `freq` is used).
        For details see https://github.com/metno/pyaerocom/pull/416.

    Raises
    ------
    TemporalResolutionError
        If `data_freq` is lower resolution than `freq`.

    Returns
    -------
    output : dict
        Dictionary with results.

    """
    if data_freq is None:
        data_freq = freq

    # input frequency is lower resolution than output frequency
    if TsType(data_freq) < TsType(freq):
        raise TemporalResolutionError(
            f"Desired input frequency {data_freq} is lower than desired output frequency {freq}"
        )

    output = {}
    if data_freq not in data or data[data_freq] is None:
        raise TemporalResolutionError(
            f"failed to compute statistics timeseries, no co-located data "
            f"available in specified base resolution {data_freq}"
        )

    coldata = data[data_freq]

    # get time index of output frequency
    to_idx = data[freq].data.time.values
    tstr = TsType(freq).to_numpy_freq()
    # list of strings of output timestamps (used below to select the
    # individual periods)
    to_idx_str = [str(x) for x in to_idx.astype(f"datetime64[{tstr}]")]
    jsdate = _get_jsdate(to_idx)

    for regid, regname in region_ids.items():
        output[regname] = {}
        try:
            subset = coldata.filter_region(region_id=regid, check_country_meta=use_country)
        except DataCoverageError:
            continue
        for i, js in enumerate(jsdate):
            per = to_idx_str[i]
            try:
                arr = ColocatedData(data=subset.data.sel(time=per))
                stats = arr.calc_statistics(use_area_weights=use_weights, drop_stats=drop_stats)
                output[regname][str(js)] = _prep_stats_json(stats)
            except DataCoverageError:
                pass

    return output


def _get_jsdate(nparr: np.ndarray):
    dt = nparr.astype("datetime64[s]")
    offs = np.datetime64("1970", "s")
    return (dt - offs).astype(int) * 1000


def _init_data_default_frequencies(coldata: ColocatedData, to_ts_types):
    """
    Compute one colocated data object for each desired statistics frequency

    Parameters
    ----------
    coldata : ColocatedData
        Initial colocated data in a certain frequency
    to_ts_types : list
        list of desired frequencies for which statistical parameters are being
        computed.

    Returns
    -------
    dict
        keys are elements of `to_ts_types`, values are corresponding
        instances of `ColocatedData` in that resolution, or None, if resolution
        is higher than original resolution of input `coldata`.
    dict
        like `data_arrs` but values are jsdate instances.

    """

    data_arrs = dict.fromkeys(to_ts_types)

    from_tst = TsType(coldata.ts_type)

    for to in to_ts_types:
        to_tst = TsType(to)
        if from_tst < to_tst:
            continue
        elif from_tst == to_tst:
            cd = coldata.copy()
        else:
            cd = coldata.resample_time(to, settings_from_meta=True, inplace=False)
        # add season coordinate for later filtering
        arr = cd.data
        arr["season"] = arr.time.dt.season
        jsdates = _get_jsdate(arr.time.values)
        arr = arr.assign_coords(jsdate=("time", jsdates))
        cd.data = arr
        data_arrs[to] = cd

    return data_arrs


def process_profile_data_for_regions(
    data: ColocatedData,
    region_id: str,
    use_country: bool,
    periods: list[str],
    seasons: list[str],
    use_meteorological_seasons: bool,
) -> dict:  # pragma: no cover
    """
    This method populates the json files in data/profiles which are use for visualization.
    Analogous to _process_map_and_scat for profile data.
    Each json file corresponds to a region or station, obs network, and variable.
    Inside the json, it is broken up by model.
    Each model has a key for "z" (the vertical dimension), "obs", and "mod"
    Each "obs" and "mod" is broken up by period.


    Args:
        data (ColocatedData): ColocatedData object for this layer
        region_id (str): Spatial subset to compute the mean profiles over
        station_name (str): Station to compute mean profiles over for period
        use_country (boolean): Passed to filter_region().
        periods (str): Year part of the temporal range to average over
        seasons (str): Sesonal part of the temporal range to average over

    Returns:
        output (dict): Dictionary to write to json
    """
    output = {"obs": {}, "mod": {}}

    for freq, coldata in data.items():
        if freq not in output["obs"]:
            output["obs"][freq] = {}
        if freq not in output["mod"]:
            output["mod"][freq] = {}

        for per in periods:
            for season in seasons:
                use_dummy = coldata is None
                perstr = f"{per}-{season}"
                if use_dummy:
                    output["obs"][freq][perstr] = np.nan
                    output["mod"][freq][perstr] = np.nan
                else:
                    try:
                        per_season_subset = _select_period_season_coldata(
                            coldata, per, season, use_meteorological_seasons
                        )

                        subset = per_season_subset.filter_region(
                            region_id=region_id, check_country_meta=use_country
                        )
                        with ignore_warnings(
                            RuntimeWarning,
                            "Mean of empty slice",
                            "All-NaN slice encountered",
                        ):
                            output["obs"][freq][perstr] = np.nanmean(subset.data[0, :, :])
                            output["mod"][freq][perstr] = np.nanmean(subset.data[1, :, :])

                    except (DataCoverageError, TemporalResolutionError) as e:
                        msg = f"Failed to access subset timeseries, and will skip. Reason was: {e}"
                        logger.info(msg)

                        output["obs"][freq][perstr] = np.nan
                        output["mod"][freq][perstr] = np.nan

    return output


def process_profile_data_for_stations(
    data: ColocatedData,
    station_name: str,
    use_country: bool,
    periods: list[str],
    seasons: list[str],
    use_meteorological_seasons: bool,
) -> dict:  # pragma: no cover
    """
    This method populates the json files in data/profiles which are use for visualization.
    Analogous to _process_map_and_scat for profile data.
    Each json file corresponds to a region, obs network, and variable.
    Inside the json, it is broken up by model.
    Each model has a key for "z" (the vertical dimension), "obs", and "mod"
    Each "obs" and "mod" is broken up by period.


    Args:
        data (ColocatedData): ColocatedData object for this layer
        region_id (str): Spatial subset to compute the mean profiles over
        station_name (str): Station to compute mean profiles over for period
        use_country (boolean): Passed to filter_region().
        periods (str): Year part of the temporal range to average over
        seasons (str): Sesonal part of the temporal range to average over

    Returns:
        output (dict): Dictionary to write to json
    """
    output = {"obs": {}, "mod": {}}

    for freq, coldata in data.items():
        if freq not in output["obs"]:
            output["obs"][freq] = {}
        if freq not in output["mod"]:
            output["mod"][freq] = {}

        for per in periods:
            for season in seasons:
                use_dummy = coldata is None
                perstr = f"{per}-{season}"
                if use_dummy:
                    output["obs"][freq][perstr] = np.nan
                    output["mod"][freq][perstr] = np.nan
                else:
                    try:
                        per_season_subset = _select_period_season_coldata(
                            coldata, per, season, use_meteorological_seasons
                        )

                        subset = per_season_subset.data[
                            :,
                            :,
                            per_season_subset.data.station_name.values
                            == station_name,  # in this case a station
                        ]  # Assumes ordering of station name matches

                        with ignore_warnings(
                            RuntimeWarning,
                            "Mean of empty slice",
                            "All-NaN slice encountered",
                        ):
                            output["obs"][freq][perstr] = np.nanmean(subset.data[0, :, :])
                            output["mod"][freq][perstr] = np.nanmean(subset.data[1, :, :])

                    except (DataCoverageError, TemporalResolutionError) as e:
                        msg = f"Failed to access subset timeseries, and will skip. Reason was: {e}"
                        logger.info(msg)

                        output["obs"][freq][perstr] = np.nan
                        output["mod"][freq][perstr] = np.nan

    return output


def _remove_less_covered(
    data: ColocatedData,
    min_yrs: int,
    sequential_yrs: bool = False,
) -> ColocatedData:
    stations = data.data.station_name.data
    data = data.copy()
    for i, s in enumerate(stations):
        years = _get_yeares(data, s)
        if sequential_yrs:
            max_yrs = _find_longest_seq_yrs(years)
        else:
            max_yrs = _find_n_yrs(years)

        if min_yrs > max_yrs:
            logger.info(f"Dropping {s}; It has only {max_yrs}")
            data.data = data.data.drop_sel(station_name=s)

    new_stations = data.data.station_name.data

    logger.info(f"Removed {len(stations) - len(new_stations)} stations")
    if len(new_stations) == 0:
        logger.warning(
            f"No stations left after removing stations with fewer than {min_yrs} years!"
        )

    return data


def _get_yeares(coldata: ColocatedData, site: str, start_yr: int = 2000) -> np.ndarray:
    found_yrs = []
    data = coldata.data.sel(station_name=site).isel(data_source=0).to_series()

    yrs = np.unique(data.index.year)
    for yr in yrs:
        if yr < start_yr:  # winter
            continue

        subset = data.loc[str(yr)]

        if len(subset) == 0:
            continue
        elif np.isnan(subset.values).all():
            continue

        d = subset.index
        valid_mask = ~np.isnan(subset.values)
        seasons = _get_unique_seasons(d[valid_mask])

        if len(seasons) == 4:
            found_yrs.append(yr)

    return np.array(found_yrs)


def _find_n_yrs(years: np.ndarray) -> int:
    return len(years)


def _find_longest_seq_yrs(years: np.ndarray) -> int:
    years = np.sort(years)
    max_yrs = 1
    yrs = 1
    for i in range(1, len(years)):
        if years[i] == years[i - 1] + 1:
            yrs += 1
        else:
            yrs = 1

        max_yrs = max(yrs, max_yrs)

    return max_yrs


def _get_yrs_from_season_yrs(season_yrs: dict) -> np.ndarray:
    yrs = []
    for y in season_yrs.keys():
        s = season_yrs[y]
        if len(list(set(s))) == 4:
            yrs.append(y)

    return np.array(yrs)


def _process_statistics_timeseries_single_region(
    reg: str,
    regnames: dict,
    data: dict[str, ColocatedData],
    main_freq: str,
    use_weights: bool,
    drop_stats: tuple,
    use_country: bool,
    input_freq: str,
    obs_name: str,
    var_name_web: str,
    vert_code: str,
    model_name: str,
    model_var: str,
):
    try:
        stats_ts = _process_statistics_timeseries(
            data=data,
            freq=main_freq,
            region_ids={reg: regnames[reg]},
            use_weights=use_weights,
            drop_stats=drop_stats,
            use_country=use_country,
            data_freq=input_freq,
        )
    except TemporalResolutionError:
        stats_ts = {}

    region = regnames[reg]
    return (stats_ts, region, obs_name, var_name_web, vert_code, model_name, model_var)


def _calculate_fairmode(
    coldata: ColocatedData,
    fairmode_statistics,  #: FairmodeStatistics,
    exp_output,  #: ExperimentOutput,
    obs_name: str,
    var_name_web: str,
    vert_code: str,
    model_name: str,
    model_var: str,
    map_meta: list[dict],
    obs_var: str = None,
    periods: tuple[str, ...] | None = None,
    seasons: tuple[str, ...] | None = None,
    use_meteorological_seasons: bool = False,
):
    for per in periods:
        results = {"ALL": {}}
        for season in seasons:
            try:
                subset = _select_period_season_coldata(
                    coldata, per, season, use_meteorological_seasons
                )
                # jsdate = subset.data.jsdate.values.tolist()
            except (DataCoverageError, TemporalResolutionError) as e:
                logger.info(f"Failed to access subset coldata: {e}")
                return results

            perstr = f"{per}-{season}"
            fm_stats = fairmode_statistics.fairmode_statistics(subset, obs_var)
            for i, station in enumerate(map_meta):
                station_name = station["station_name"]
                region = station["region"][0]
                if region not in results:
                    results[region] = {}

                if perstr not in results[region]:
                    results[region][perstr] = {}
                if perstr not in results["ALL"]:
                    results["ALL"][perstr] = {}

                results[region][perstr][station_name] = fm_stats[station_name]
                results["ALL"][perstr][station_name] = fm_stats[station_name]

        for reg in results:
            fairmode_statistics.save_fairmode_stats(
                exp_output,
                results,
                obs_name,
                var_name_web,
                vert_code,
                model_name,
                model_var,
                per,
                reg,
            )



from __future__ import annotations


from tests.fixtures.data_access import TEST_DATA
from pyaerocom.io.mscw_ctm.reader import ReadMscwCtm
from pyaerocom import GriddedData

import pytest

EMEP_DATA_PATH = TEST_DATA["MODELS"].path / "EMEP_cities"
UEMEP_DATA_PATH = TEST_DATA["MODELS"].path / "uEMEP_cities"

EMEP_FILE_PATHS = [
    EMEP_DATA_PATH / "Amsterdam",
    EMEP_DATA_PATH / "Berlin",
]

EMEP_FILE_PATHS_SPLIT = [
    EMEP_DATA_PATH / "Berlin_split_1",
    EMEP_DATA_PATH / "Berlin_split_2",
]

UEMEP_FILE_PATHS = [
    UEMEP_DATA_PATH / "Bordeaux",
    UEMEP_DATA_PATH / "Lyon",
]


@pytest.fixture
def cities_data() -> dict[str, list[GriddedData]]:
    data_dict = {}

    data_dict["uEMEP"] = [
        ReadMscwCtm("uemep", str(ddir)).read_var("concpm25", ts_type="hourly")
        for ddir in UEMEP_FILE_PATHS
    ]
    data_dict["EMEP"] = [
        ReadMscwCtm("emep", str(ddir)).read_var("concpm25", ts_type="monthly")
        for ddir in EMEP_FILE_PATHS
    ]

    data_dict["EMEP_split"] = [
        ReadMscwCtm("emep", str(ddir)).read_var("concpm25", ts_type="monthly")
        for ddir in EMEP_FILE_PATHS_SPLIT
    ]

    return data_dict



import pytest
from copy import deepcopy

from pyaerocom.griddeddata_container import GriddedDataContainer, GriddedDataContainerException
from pyaerocom import GriddedData


def test___init__():
    data_id = "test_id"
    data = GriddedDataContainer(data_id)
    assert data.data_id == data_id


def test_fixture(cities_data):
    for model in ["uEMEP", "EMEP"]:
        data = cities_data[model]
        for city in data:
            assert isinstance(city, GriddedData)


@pytest.mark.parametrize(
    "model",
    [
        "uEMEP",
        "EMEP",
        "EMEP_split",
    ],
)
def test__initiate(cities_data, model):
    data_id = "test_id"
    mg = GriddedDataContainer(data_id)

    gd = cities_data[model][0]
    mg._initiate(gd)

    assert mg.proj_info == gd.proj_info
    assert mg.start == gd.start


@pytest.mark.parametrize(
    "model",
    [
        "uEMEP",
        "EMEP",
    ],
)
def test_add_griddeddata(cities_data, model):
    data_id = "test_id"
    mg = GriddedDataContainer(data_id)

    for gd in cities_data[model]:
        mg.add_griddeddata(gd)

    assert len(mg.children) == len(cities_data[model])


def test_get_xyranges(cities_data):
    correct_xs = [[3462625, 3495375], [3908625, 3935875]]
    correct_ys = [[2462125, 2486375], [2516625, 2541375]]
    data_id = "test_id"
    mg = GriddedDataContainer(data_id)

    for gd in cities_data["uEMEP"]:
        mg.add_griddeddata(gd)

    xs, ys = mg.get_xyranges()

    for ls, c_ls in zip(sorted(xs), sorted(correct_xs)):
        for _l, c_l in zip(ls, c_ls):
            assert abs(_l - c_l) < 1e-5

    for ls, c_ls in zip(sorted(ys), sorted(correct_ys)):
        for _l, c_l in zip(ls, c_ls):
            assert abs(_l - c_l) < 1e-5


def test_regrid(cities_data):
    data_id = "test_id"
    mg = GriddedDataContainer(data_id)

    for gd in cities_data["EMEP"]:
        mg.add_griddeddata(gd)

    lat_res = 0.001
    lon_res = 0.001

    for child in mg.children:
        assert child.lat_res != lat_res
        assert child.lon_res != lon_res

    mg.regrid(lat_res_deg=lat_res, lon_res_deg=lon_res)

    for child in mg.children:
        assert child.lat_res == pytest.approx(lat_res, rel=1e-6)
        assert child.lon_res == pytest.approx(lon_res, rel=1e-6)


def test_get_tiles(cities_data):
    data_id = "test_id"
    mg = GriddedDataContainer(data_id)

    for gd in cities_data["uEMEP"]:
        mg.add_griddeddata(gd)

    tiles, xranges, yranges = mg.get_tiles()

    assert len(tiles) == 2
    assert len(tiles) == len(xranges)
    assert len(xranges) == len(yranges)


# Tests for error handling


def test_regrid_fail_due_to_projection(cities_data):
    data_id = "test_id"
    mg = GriddedDataContainer(data_id)

    for gd in cities_data["uEMEP"]:
        mg.add_griddeddata(gd)

    lat_res = 0.001
    lon_res = 0.001

    with pytest.raises(
        GriddedDataContainerException,
        match="Could not regrid, since the data is projected. Please use data with latlon instead",
    ):
        mg.regrid(lat_res_deg=lat_res, lon_res_deg=lon_res)


def test_add_griddeddata_different_model(cities_data):
    data_id = "test_id"
    mg = GriddedDataContainer(data_id)

    emep = cities_data["EMEP"][0]
    uemep = cities_data["uEMEP"][0]

    mg.add_griddeddata(emep)

    with pytest.raises(
        GriddedDataContainerException,
        match="Proj info from added griddeddata and existing proj info do not have the same type*",
    ):
        mg.add_griddeddata(uemep)


@pytest.mark.parametrize(
    "attribute,attr_value",
    [
        ["var_name", "concpm10"],
        ["ts_type", "daily"],
        ["units", "1"],
    ],
)
def test_add_griddeddata_errors(cities_data, attribute, attr_value):
    data_id = "test_id"
    mg = GriddedDataContainer(data_id)

    emep = cities_data["EMEP"][0]
    emep2 = deepcopy(cities_data["EMEP"][1])

    setattr(emep2, attribute, attr_value)

    mg.add_griddeddata(emep)

    with pytest.raises(
        GriddedDataContainerException,
        match=f"{attribute} of added griddeddata*",
    ):
        mg.add_griddeddata(emep2)


@pytest.mark.skip(reason="lat_res currently not working for fine resolution test data")
def test_only_one_child(cities_data):
    data_id = "test_id"
    mg = GriddedDataContainer(data_id)

    mg.add_griddeddata(cities_data["uEMEP"][0])

    lat_res = mg.lat_res

    assert float(lat_res) == pytest.approx(0.00032, rel=0.01)

    mg.add_griddeddata(cities_data["uEMEP"][1])
    with pytest.raises(
        NotImplementedError,
        match="lat_res is not implemented for cases*",
    ):
        lat_res = mg.lat_res



import xarray
from pyproj import CRS, transformer


class ProjectionInformationException(Exception):
    pass


class ProjectionInformation:
    def __init__(self):
        self._crs = None
        self._x_axis = None
        self._y_axis = None
        self._units = None

    @property
    def x_axis(self):
        return self._x_axis

    @property
    def y_axis(self):
        return self._y_axis

    def to_proj(self, latitude, longitude):
        """convert latitude and longitude to x and y

        :param latitude: latitude values
        :param longitude: longitude values
        :return: tuple of x and y coordinates
        """
        trans = transformer.Transformer.from_crs(4326, self._crs)
        return trans.transform(latitude, longitude)

    def to_latlon(self, x, y):
        """convert x and y to latitude and longitude

        :param x: x values
        :param y: y values
        :return: tuple of latitude and longitude coordinates
        """
        trans = transformer.Transformer.from_crs(self._crs, 4326)
        return trans.transform(x, y)

    @staticmethod
    def from_xarray(ds: xarray.Dataset, var: str):
        """initialize the ProjectionInformation from an xarray variable

        returns None if no projection exists for the variable or ProjectionInformation
        """
        da = ds[var]
        if "grid_mapping" not in da.attrs:
            return None
        pi = ProjectionInformation()
        grid_mapping = ds[da.attrs["grid_mapping"]]
        pi._crs = CRS.from_cf(grid_mapping.attrs)

        for c in da.coords:
            if len(da.coords[c].dims) != 1:
                continue
            if "standard_name" in da.coords[c].attrs:
                if da.coords[c].standard_name in (
                    "longitude",
                    "grid_longitude",
                    "projection_x_coordinate",
                ):
                    pi._x_axis = c
                    break
            if "axis" in da.coords[c].attrs:
                if da.coords[c].axis in ("x", "X"):
                    pi._x_axis = c
                    break
            units = ds.coords[c].attrs.get("units")
            if units in [
                "degrees_east",
                "degree_east",
                "degree_E",
                "degrees_E",
                "degreeE",
                "degreesE",
            ]:
                pi._x_axis = c
                break
        for c in da.coords:
            if len(da.coords[c].dims) != 1:
                continue
            if "standard_name" in da.coords[c].attrs:
                if da.coords[c].standard_name in (
                    "latitude",
                    "grid_latitude",
                    "projection_y_coordinate",
                ):
                    pi._y_axis = c
                    break
            if "axis" in da.coords[c].attrs:
                if da.coords[c].axis in ("y", "Y"):
                    pi._y_axis = c
                    break
            units = ds.coords[c].attrs.get("units")
            if units in [
                "degrees_north",
                "degree_north",
                "degree_N",
                "degrees_N",
                "degreeN",
                "degreesN",
            ]:
                pi._y_axis = c
                break
        if pi._x_axis is None or pi._y_axis is None:
            raise ProjectionInformationException(f"no x or y axis found for variable '{var}'")
        pi._units = da.coords[pi._y_axis].units
        return pi

    def __eq__(self, other):
        if not isinstance(other, type(self)):
            return False
        return self.__dict__ == other.__dict__



# isort:skip_file
from importlib import metadata

import xarray

from ._logging import change_verbosity
from ._warnings import ignore_basemap_warning, ignore_earth_radius_warning

__version__ = metadata.version(__package__)

import iris
import logging


logger = logging.getLogger(__name__)

# Enable new iris functionality to suppress deprecation warning.
# https://scitools-iris.readthedocs.io/en/latest/generated/api/iris.html#iris.FUTURE
iris.FUTURE.save_split_attrs = True

try:
    iris.FUTURE.date_microseconds = True
except AttributeError:
    # Old iris version that doesn't support this override. Use old behaviour.
    pass

from .config_reader import ConfigReader

try:
    # Opt in to new defaults introduced by xarray 2025.08.0
    # https://docs.xarray.dev/en/stable/whats-new.html#v2025-08-0-august-14-2025
    xarray.set_options(
        use_new_combine_kwarg_defaults=True, netcdf_engine_order=["netcdf4", "scipy", "h5netcdf"]
    )
except ValueError:
    # xarray < 2025.08.0 so doesn't apply.
    pass


_has_warned_const_deprecation = False


def __getattr__(key: str):
    global _has_warned_const_deprecation
    # Ensures that get_instance() is called each time const is accessed, which
    # should make const mockable in tests by mocking get_instance on Config class.
    if key == "const":
        if not _has_warned_const_deprecation:
            logger.warning(
                "Use of pyaerocom.const is deprecated. Please use Config.get_instance() or pyaerocom.config instead."
            )
            _has_warned_const_deprecation = True
        return ConfigReader.get_instance()
    if key == "config":
        return ConfigReader.get_instance()
    raise AttributeError(f"Module '{__name__}' has no attribute '{key}'")


ignore_basemap_warning()
ignore_earth_radius_warning()

# Sub-packages
from . import io

from . import scripts

# Imports
from . import obs_io
from . import metastandards
from . import vertical_profile
from . import mathutils
from . import geodesy
from . import region_defs
from . import region

# from . import vert_coords
from . import stationdata
from . import griddeddata
from . import ungriddeddata
from . import colocation
from . import var_groups
from . import combine_vardata_ungridded
from . import helpers_landsea_masks
from . import helpers
from . import trends_helpers
from . import trends_engine

# custom toplevel classes
from .variable import Variable
from .region import Region
from .vertical_profile import VerticalProfile
from .stationdata import StationData
from .griddeddata import GriddedData
from .griddeddata_container import GriddedDataContainer
from .ungriddeddata import UngriddedData
from .ungriddeddata_structured import UngriddedDataStructured
from .colocation.colocated_data import ColocatedData
from .colocation.colocator import Colocator
from .colocation.colocation_setup import ColocationSetup
from .filter import Filter
from .units.datetime import TsType
from .time_resampler import TimeResampler
from .io.helpers import search_data_dir_aerocom
from .variable_helpers import get_variable
from .utils import create_varinfo_table

from . import aeroval

from .sample_data_access import download_minimal_dataset



import pytest

from pyaerocom._lowlevel_helpers import (
    check_dir_access,
    invalid_input_err_str,
    sort_dict_by_name,
    str_underline,
    BrowseDict,
)


@pytest.mark.parametrize("title", ["", "Bla", "Hello"])
@pytest.mark.parametrize("indent", [0, 4, 10])
def test_str_underline(title: str, indent: int):
    lines = str_underline(title, indent).split("\n")
    assert len(lines) == 2
    assert len(lines[0]) == len(lines[1]) == len(title) + indent
    assert lines[0].endswith(title)
    assert lines[1].endswith("-" * len(title))
    assert lines[0][:indent] == lines[1][:indent] == " " * indent


def test_invalid_input_err_str():
    st = invalid_input_err_str("bla", "42", (42, 43))
    assert st == "Invalid input for bla (42), choose from (42, 43)"


@pytest.mark.parametrize("dir,val", [(".", True), ("/bla/blub", False), (42, False)])
def test_check_dir_access(dir, val):
    assert check_dir_access(dir) == val


@pytest.mark.parametrize("input", [{"b": 1, "a": 2, "kl": 42}])
@pytest.mark.parametrize(
    "pref_list,output_keys",
    [
        ([], ["a", "b", "kl"]),
        (["blaaa"], ["a", "b", "kl"]),
        (["kl"], ["kl", "a", "b"]),
        (["kl", "b"], ["kl", "b", "a"]),
    ],
)
def test_sort_dict_by_name(input, pref_list, output_keys):
    sorted = sort_dict_by_name(input, pref_list)
    assert list(sorted.keys()) == output_keys


def test_BrowseDict():
    bd = BrowseDict(key=42)
    assert bd["key"] == bd.key == 42


def test_BrowseDict_assign():
    bd = BrowseDict()

    bd.key = "test"

    assert bd["key"] == bd.key == "test"


def test_BrowseDict_forbidden_keys():
    bd = BrowseDict()
    bd.FORBIDDEN_KEYS.append("forbidden")

    with pytest.raises(KeyError):
        bd["forbidden"] = "test"



"""
Config file for AeroCom PhaseIII test project
"""
# Imported and updated from: https://gitlab.met.no/aeroval/config/-/blob/master/eval_py/gridded_io_aux.py

from pyaerocom.io.aux_read_cubes import (
    add_cubes,
    subtract_cubes,
    divide_cubes,
    multiply_cubes,
    compute_angstrom_coeff_cubes,
    mmr_to_vmr_cube,
    conc_from_vmr_STP,
)
from pyaerocom.units.molecular_mass import MolecularMass


M_N = float(MolecularMass("N"))
M_O = float(MolecularMass("O"))
M_H = float(MolecularMass("H"))


def calc_concnh3(concnh3):  # pragma: no cover
    in_ts_type = concnh3.ts_type

    concNnh3 = concnh3 * (M_N / (M_N + M_H * 3))
    concNnh3.units = "ug N m-3"

    concNnh3.attributes["ts_type"] = in_ts_type

    return concNnh3


def calc_concnh4(concnh4):  # pragma: no cover
    if concnh4.units == "ug m-3" or concnh4.units == "ug/m**3":
        concnh4.units = "ug/m3"
    assert concnh4.units == "ug/m3"

    in_ts_type = concnh4.ts_type
    concnh4 = concnh4.cube
    concnh4 *= M_N / (M_N + M_H * 4)
    concnh4.units = "ug N m-3"

    concnh4.attributes["ts_type"] = in_ts_type

    return concnh4


def calc_conchno3(vmrhno3):  # pragma: no cover
    if vmrhno3.units == "1e-9":
        vmrhno3.units == "ppb"
    assert vmrhno3.units == "ppb"

    in_ts_type = vmrhno3.ts_type
    conchno3 = conc_from_vmr_STP(vmrhno3.cube)
    conchno3.units = "ug/m3"
    conchno3 *= M_N / (M_H + M_N + M_O * 3)
    conchno3.attributes["ts_type"] = in_ts_type
    conchno3.units = "ug N m-3"

    return conchno3


def calc_fine_concno310(concno3f):  # pragma: no cover
    return calc_concno310(concno3f=concno3f, concno3c=None)


def calc_concno310(concno3c, concno3f):  # pragma: no cover
    if concno3c is not None:
        if concno3c.units == "ug m-3" or concno3c.units == "ug/m**3":
            concno3c.units = "ug/m3"
        assert concno3c.units == "ug/m3"
    assert concno3f.units == "ug/m3"

    in_ts_type = concno3f.ts_type
    if concno3c is not None:
        concno310 = add_cubes(concno3f.cube, concno3c.cube)
    else:
        concno310 = concno3f.cube

    concno310 *= M_N / (M_N + M_O * 3)
    concno310.attributes["ts_type"] = in_ts_type
    concno310.units = "ug N m-3"
    return concno310


def calc_concno325(concno3f):  # pragma: no cover
    assert concno3f.units == "ug/m3"
    in_ts_type = concno3f.ts_type
    concno325 = concno3f.cube
    concno325 *= M_N / (M_N + M_O * 3)

    concno325.attributes["ts_type"] = in_ts_type
    concno325.units = "ug N m-3"
    return concno325


def calc_fine_conctno3(concno3f, vmrhno3):  # pragma: no cover
    return calc_conctno3(concno3f=concno3f, concno3c=None, vmrhno3=vmrhno3)


def calc_conctno3(concno3c, concno3f, vmrhno3):  # pragma: no cover
    if concno3c is not None:
        if concno3c.units == "ug m-3" or concno3c.units == "ug/m**3":
            concno3c.units = "ug/m3"
        assert concno3c.units == "ug/m3"

    if vmrhno3.units == "1e-9":
        vmrhno3.units == "ppb"
    assert concno3f.units == "ug/m3"
    assert vmrhno3.units == "ppb"

    in_ts_type = vmrhno3.ts_type
    if concno3c is not None:
        concno3 = add_cubes(concno3f.cube, concno3c.cube)
    else:
        concno3 = concno3f.cube

    conchno3 = conc_from_vmr_STP(vmrhno3.cube)
    conchno3.units = "ug/m3"
    concno3 *= M_N / (M_N + M_O * 3)
    conchno3 *= M_N / (M_H + M_N + M_O * 3)
    conctno3 = add_cubes(concno3, conchno3)
    conctno3.attributes["ts_type"] = in_ts_type
    conctno3.units = "ug N m-3"
    return conctno3


def calc_conctnh(concnh4, vmrnh3):  # pragma: no cover
    if concnh4.units == "ug m-3" or concnh4.units == "ug/m**3":
        concnh4.units = "ug/m3"
    if vmrnh3.units == "1e-9":
        vmrnh3.units = "ppb"
    assert concnh4.units == "ug/m3"
    assert vmrnh3.units == "ppb"

    concnh3 = conc_from_vmr_STP(vmrnh3.cube)
    concnh3.units = "ug/m3"
    concnh3 *= M_N / (M_N + M_H * 3)
    in_ts_type = concnh4.ts_type
    concnh4 = concnh4.cube
    concnh4 *= M_N / (M_N + M_H * 4)
    conctnh = add_cubes(concnh3, concnh4)
    conctnh.attributes["ts_type"] = in_ts_type
    conctnh.units = "ug N m-3"
    return conctnh


def calc_aod_from_species_contributions(*gridded_objects):  # pragma: no cover
    data = gridded_objects[0].cube

    assert str(data.units) == "1"

    for obj in gridded_objects[1:]:
        assert str(obj.units) == "1"
        data = add_cubes(data, obj.cube)

    return data


FUNS = {
    "add_cubes": add_cubes,
    "subtract_cubes": subtract_cubes,
    "divide_cubes": divide_cubes,
    "multiply_cubes": multiply_cubes,
    "calc_ae": compute_angstrom_coeff_cubes,
    "calc_conctno3": calc_conctno3,
    "calc_fine_conctno3": calc_fine_conctno3,
    "calc_conctnh": calc_conctnh,
    "calc_concnh3": calc_concnh3,
    "calc_concnh4": calc_concnh4,
    "calc_conchno3": calc_conchno3,
    "calc_concno310": calc_concno310,
    "calc_fine_concno310": calc_fine_concno310,
    "calc_concno325": calc_concno325,
    "calc_aod_from_species_contributions": calc_aod_from_species_contributions,
    "mmr_to_vmr": mmr_to_vmr_cube,
}



from .data_access import DataForTests
from pyaerocom.io import PyaroConfig, ReadPyaro
import pytest

OBS_PATH = DataForTests("obsdata/LCS/2019").path


@pytest.fixture
def lcs_data():
    reader_id = "lcsreader"
    config = PyaroConfig(
        name="test-lcs",
        reader_id=reader_id,
        filename_or_obj_or_url=str(OBS_PATH),
        filters={},
        network="SC",
        min_quality=2,
        name_map={
            "PM25": "concpm25",
        },
    )

    reader = ReadPyaro(config)

    return reader.read(vars_to_retrieve=["concpm25"])
